/* 
** Copyright 1994 by Miron Livny, and Mike Litzkow
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:  Mike Litzkow
**
*/ 

/*
This program computes for approximately 10 hours, then exits with
status 0.  First the program grows its data size to match a command
line argument.
	Every 10 seconds the total running seconds is printed to stdout.
	Every minute the resource usage is printed to stdout.
	Every minute the current time is printer onto a "timestamp" file.
	Every minute the avgv and environ are checked against a saved buffer.
The program just counts in a loop, so the actual time may vary if the
CPU is faster or slower than expected.

Command line arguments
	1. Size the program should grow to (in megabytes ).
	2. Name of the "timestamp" file.
*/

#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/resource.h>

#if defined(AIX32)
#	include <time.h>
#endif


extern int CkptInProgress;

long data_start_addr();
char	*sbrk();

void save_strings( int argc, char *argv[], char **envp, char *buf );
void check_strings( int argc, char *argv[], char **envp, char *buf );
char *append( char *str, char *buf );
void report_rusage( int who, char *str );
void check_size();

extern char	**environ;

#define MEG (1024 * 1024)
char	Buf1[4096], Buf2[4096];

main( argc, argv)
int		argc;
char	*argv[];
{
	int		count = 36000;
	int		size;
	int		i;
#if defined(DEBUGGING)
	int		wait_for_me = 1;

	while( wait_for_me )
		;
#endif

	save_strings( argc, argv, environ, Buf1 );

	if( argc != 3 ) {
		fprintf( stderr, "Usage: %s megs filename\n", argv[0] );
		exit( 1 );
	}

	size = atoi( argv[1] );

	open_timestamp_file( argv[2] );

	check_size();
	for( i=0; i<size; i++ ) {
		(void)sbrk( MEG );
		check_size();
	}

	do_loop( count, argc, argv, environ );
	printf( "Normal End of Job\n" );

	exit( 0 );
}

do_loop( count, argc, argv, environ )
int		count;
int		argc;
char	*argv[];
char	**environ;
{
	int		i;

	for( i=0; i<count; i++ ) {
		if( i % 60 == 0 ) {
			printf( "\n" );
			report_rusage( RUSAGE_SELF, "Self" );
			timestamp();
			memset( Buf2, '\0', sizeof(Buf2) );
			save_strings( argc, argv, environ, Buf2 );
			if( memcmp(Buf1,Buf2,sizeof(Buf1)) == 0 ) {
				printf( "Check %d: Buffers are equal\n", i/60 );
				fflush( stdout );
			} else {
				printf( "Check %d: Buffers differ\n", i );
				gen_core();
			}
		}
		waste_a_second();
		if( i % 10 == 0 ) {
			printf( "%d ", i );
			fflush( stdout );
		}
	}
	printf( "\n" );
	report_rusage( RUSAGE_SELF, "Self" );
}


#define UNIT 10000

#if defined(ALPHA)
#	define LIM (800 * UNIT)
#elif defined(AIX32)
#	define LIM (225 * UNIT)
#elif defined(SPARC)
#	define LIM (260 * UNIT)
#elif defined(ULTRIX42)
#	define LIM (170 * UNIT)
#elif defined(HPPAR)
#	define LIM (260 * UNIT)
#endif

waste_a_second()
{
	int		i;

	for( i=0; i<LIM; i++ )
		;
}

usage()
{
	printf( "Usage: loop count\n" );
	exit( 1 );
}

void
report_rusage( who, str )
int		who;
char	*str;
{
	struct rusage	buf;

#if defined(HPUX9)
	printf( "Resource usage reporting not supported by HP-UX\n" );
#elif defined(OSF1) || defined(SUNOS41)
	printf( "Resource usage reporting not supported for V5 Condor yet\n" );
#else
	if( getrusage(who,&buf) < 0 ) {
		printf( "getrusage(%d) failed\n", who );
		return;
	}

	printf( "Resource usage for %s:\n", str );
	printf( "\tUserTime = %d.%06d\n",
		buf.ru_utime.tv_sec, buf.ru_utime.tv_usec );
	printf( "\tSystemTime = %d.%06d\n",
		buf.ru_stime.tv_sec, buf.ru_stime.tv_usec );
#endif
}

int	Tfd;

open_timestamp_file( name )
char	*name;
{
	if( (Tfd = open(name,O_WRONLY|O_CREAT|O_TRUNC,0644)) < 0 ) {
		perror( name );
		exit( 1 );
	}
}

timestamp()
{
	time_t	now;
	char	*time_string;

	if( lseek(Tfd,0,2) < 0 ) {
		perror( "lseek" );
		exit( 1 );
	}

	now = time(0);
	time_string = ctime( &now );
	write( Tfd, time_string, strlen(time_string) );
}

void
save_strings( int argc, char *argv[], char **envp, char *buf )
{
	int		i;

	for( i=0; i<argc; i++ ) {
		buf = append( argv[i], buf );
	}
	for( envp++; *envp; envp++ ) {
		buf = append( *envp, buf );
	}
}

char *
append( char *str, char *buf )
{
	strcpy( buf, str );
	while( *buf ) {
		buf++;
	}
	return buf + 1;
}

void
check_size()
{
	printf(
		"VM size is now %.1f\n",
		(float)( ((long)sbrk(0) - data_start_addr()) / MEG )
	);
}

gen_core()
{
	char	*foo;

	foo = (char *)0;
	*foo = 'a';
	exit( 99 );
}

#if defined(OSF1)
#	include <sys/addrconf.h>
	static struct addressconf AddrTab[ AC_N_AREAS ];
	const int AddrTabSize = sizeof( AddrTab );

	void
	init_addr_tab()
	{
		static int initialized = FALSE;

		if( initialized ) {
			return;
		}

		if( getaddressconf(AddrTab,AddrTabSize) != AddrTabSize ) {
			perror( "getaddressconf" );
			exit( 1 );
		}
	}

	/*
	  Return starting address of the data segment
	*/
	long
	data_start_addr()
	{
		init_addr_tab();
		return (long)AddrTab[ AC_DATA ].ac_base;
	}

#elif defined(SUNOS41)
#	include <sys/param.h>
	extern int etext;
	long
	data_start_addr()
	{
		return ((long)&etext + DATA_ALIGN - 1) / DATA_ALIGN * DATA_ALIGN;
	}

#elif defined(HPUX9)
	/*
	  Return starting address of the data segment
	*/
	extern int __data_start;
	long
	data_start_addr()
	{
		return (long)&__data_start;
	}

#elif defined(AIX32)
	extern int _data;
	long
	data_start_addr()
	{
		return (long)&_data;
	}
#endif
