#include	<stdio.h>
#include	<signal.h>

#define SENT 1
#define RECEIVED 2

extern int SetSyscalls(int);
int sig_array[ NSIG + 1 ];

handler(sig)
int	sig;
{
	printf("User signal handler: got signal %d\n", sig );
	mark( sig, RECEIVED );
	printf("Leaving user signal handler\n");
}

main()
{
	int	sig;


		/* First we send ourselves every reasonable signal,
		** but we keep them all blocked. */
	for( sig=1; sig < NSIG; sig++ ) {
		switch( sig ) {
				
				/* Special Condor Signals */
			case SIGQUIT:
			case SIGUSR1:
			case SIGTSTP: 
				continue;

				/* Can't be caught or ignored */
			case SIGKILL:
			case SIGSTOP:
				continue;

				/* Wierd ones ... */
			case SIGCONT:
			case SIGTRAP:
				continue;

		}

		catch_sig( sig, handler );
		sigblock( sigmask(sig) );
		kill( getpid(), sig ); 
		mark( sig, SENT );
	}

		/* Checkpoint and restart */
	/* display_state(); */
	ckpt();

		/* Ok, now unblock them all */
	printf( "About to unblock all signals\n" );
	fflush( stdout );
	/* display_state(); */
	sigsetmask( 0 );  

		/* Make sure we got evey one */
	if( check_sigs_received() ) {
		exit( 0 );
	} else {
		exit( 1 );
	}
}


mark( sig, val )
{
	/*
	printf( "mark(%d,%d)\n", sig, val );
	*/
	switch( val ) {
		case SENT:
			if( sig_array[ sig ] != 0 ) {
				fprintf( stderr, "Signal %d sent twice\n", sig );
				abort();
			}
			break;
		case RECEIVED:
			if( sig_array[ sig ] == 0 ) {
				fprintf( stderr, "Signal %d received, but never sent\n", sig );
				abort();
			}
			if( sig_array[ sig ] == RECEIVED ) {
				fprintf( stderr, "Signal %d received twice\n", sig );
				abort();
			}
			break;
	}
	sig_array[ sig ] = val;
}

int
check_sigs_received()
{
	int		i;
	int		ok = 1;

	printf("About to check to see if all signals received....\n");
	for( i=1; i<NSIG; i++ ) {
		if( sig_array[i] == SENT ) {
			printf( "Signal %d sent, but never received\n", i );
			ok = 0;
		}
	}
	if( ok ) {
		printf( "All signals received\n" );
	}
	return ok;
}

catch_sig( sig, handler )
int		sig;
void	(*handler)();
{
	struct sigaction vec;

	vec.sa_handler = handler;
	sigfillset(&vec.sa_mask);
	vec.sa_flags = 0;
	if ( sigaction(sig,&vec,NULL) < 0 ) {
		perror("sigaction");
		exit(1);
	}
}

#ifdef 0
catch_sig( sig, handler )
int		sig;
void	(*handler)();
{
	struct sigvec	vec;

	vec.sv_handler = handler;
	vec.sv_mask = -1; 
	vec.sv_flags = 0;
#ifdef HPUX9
	if( sigvector(sig,&vec,0) < 0 ) {
#else
	if( sigvec(sig,&vec,0) < 0 ) {
#endif
		perror( "sigvec" );
		exit( 1 );
	}
}
#endif

#ifdef 0
display_state()
{
	sigset_t	pending;
	sigset_t	mask;
	struct sigaction oact;
		int		i;

	sigpending( &pending );
	sigprocmask( SIG_BLOCK, NULL, &mask );
	for( i=1; i<NSIG; i++ ) {
		printf( "Sig %d: ", i );
		if( sigismember(&pending,i) ) {
			printf( "Pending " );
		} else {
			printf( "NOT Pending " );
		}
		sigaction( i, NULL, &oact );
		printf( "Handler = 0x%p ", oact.sa_handler );
		if( sigismember(&mask,i) ) {
			printf( "Blocked " );
		} else {
			printf( "NOT Blocked " );
		}
		printf( "\n" );
	}
}
#endif
