#define _POSIX_SOURCE

#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <time.h>

#define TRUE 1
#define FALSE 0

void display_stat( const char *name, struct stat *buf );
void do_it( const char *name );
int stat_eq( struct stat *buf1, struct stat *buf2 );

main( int argc, char *argv[] )
{
	int		i;

	for( i=1; i<argc; i++ ) {
		do_it( argv[i] );
	}

	printf( "Normal End Of Job\n" );
	exit( 0 );
}

void
do_it( const char *name )
{
	struct stat	buf1;
	struct stat	buf2;
	int		fd;

		/* Get the data with stat() */
	memset( &buf1, '\0', sizeof(buf1) );
	if( stat(name,&buf1) < 0 ) {
		perror( "stat" );
		exit( __LINE__ );
	}

		/* Get the data with fstat() */
	memset( &buf2, '\0', sizeof(buf2) );
	if( (fd=open(name,O_RDONLY)) < 0 ) {
		perror( name );
		exit( __LINE__ );
	}
	if( fstat(fd,&buf2) < 0 ) {
		perror( "stat" );
		exit( __LINE__ );
	}
	close( fd );

	if( !stat_eq(&buf1,&buf2) ) {
		fprintf( stderr, "Error, stat buffers differ\n" );
		exit( __LINE__ );
	}

	display_stat( name, &buf1 );
}

void
display_stat( const char *name, struct stat *buf )
{
	printf( "%s:\n", name );
	printf( "  st_mode = 0%o\n", buf->st_mode & 0777 );
	printf( "  st_ino = %d\n", buf->st_ino );
	printf( "  st_dev = %d\n", buf->st_dev );
	printf( "  st_nlink = %d\n", buf->st_nlink );
	printf( "  st_uid = %d\n", buf->st_uid );
	printf( "  st_gid = %d\n", buf->st_gid );
	printf( "  st_size = %d\n", buf->st_size );
	printf( "  st_atime = %s", ctime(&buf->st_atime) );
	printf( "  st_mtime = %s", ctime(&buf->st_mtime) );
	printf( "  st_ctime = %s", ctime(&buf->st_ctime) );
}

#define COMP(field) 						\
	if( buf1->field != buf2->field ) {		\
		printf( "%s differs\n", #field );	\
		return FALSE;						\
	}

int
stat_eq( struct stat *buf1, struct stat *buf2 )
{
	COMP(st_mode)
	COMP(st_ino)
	COMP(st_dev)
	COMP(st_nlink)
	COMP(st_uid)
	COMP(st_gid)
	COMP(st_size)
	COMP(st_atime)
	COMP(st_mtime)
	COMP(st_ctime)

	return TRUE;
}
