/* 
** Copyright 1986, 1987, 1988, 1989, 1990, 1991 by the Condor Design Team
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the Condor Design Team not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the Condor
** Design Team make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE CONDOR DESIGN TEAM DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE CONDOR DESIGN TEAM BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:   Dhrubajyoti Borthakur
** 	         University of Wisconsin, Computer Sciences Dept.
** 
*/ 
#define _POSIX_SOURCE  // FIX

#include <stdio.h>

#if defined(IRIX331)
#define __EXTENSIONS__
#include <signal.h>
#undef __EXTENSIONS__
#define BADSIG SIG_ERR
#else
#include <signal.h>
#endif

#include <errno.h>
#include <pwd.h>
#include <netdb.h>
#include <string.h>
#if 0
#include <strings.h>
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include "condor_fix_timeval.h"  // FIX
#include <sys/resource.h>
#include <sys/wait.h>
#include "condor_constants.h"

#include "condor_common.h"

#include "condor_debug.h"
#include "condor_expressions.h"
#include "condor_config.h"
#include "condor_uid.h"
#include "condor_mach_status.h"
#include "master.h"
#include "file_lock.h"

#if defined(HPUX9)
#include "fake_flock.h"
#define _BSD
#endif

typedef void (*SIGNAL_HANDLER)();
void install_sig_handler( int, SIGNAL_HANDLER );

// these are defined in master.C
extern int		NotFlag;
extern int 		RestartsPerHour;
extern int 		MasterLockFD;
extern char *		_FileName_ ;
extern FileLock*	MasterLock;

extern int     collector_runs_here();
extern int     negotiator_runs_here();
extern time_t  GetTimeStamp(char* file);
extern void    do_killpg(int, int);
extern void    give_up(char* );
extern int 	   NewExecutable(char* file, time_t* tsp);
int		hourly_housekeeping(void);
extern "C"
{
#if	defined(SUNOS41)
	extern	int vfork();
#endif 
	extern	int		event_mgr();
}

// to add a new process as a condor deamon, just add one line in 
// the structure below. The first elemnt is the string that is 
// looked for in the config file for the executable, and the
// second element is the parameter looked for in the confit
// file for the name of the corresponding log file.If no log
// file need be there, then put a zero in the second column.
// The third parameter is the name of a condor_config variable
// that is ckecked before the process is created. If it is zero, 
// then the process is created always. If it is a valid name,
// this name shud be set to true in the condor_config file for the
// process to be created.

// The PVMD deamon do not have any log files as the second column is 0.
// The PVMD is started only if the values of PVMD_RUNS_HERE is
// set to TRUE in the config file
// make sure that the master does not start itself : set the flag off

// if this machine is a world-machine, then the standard schedd and
// startd are not started.

deamon Deamons::deamon_ptr[] = 
	{
	{"MASTER", 		"MASTER_LOG", 				      0,0,0,FALSE,0,0,0},
	{"COLLECTOR",	"COLLECTOR_LOG",			      0,0,0,TRUE, 0,0,0},
	{"NEGOTIATOR",	"NEGOTIATOR_LOG", 			      0,0,0,TRUE, 0,0,0},
	{"STARTD",		"STARTD_LOG" , 				      0,0,0,TRUE, 0,0,0},
	{"SCHEDD",		"SCHEDD_LOG", 				      0,0,0,TRUE, 0,0,0},
	{"KBDD",		"KBDD_LOG", 	      "X_RUNS_HERE",0,0,TRUE, 0,0,0},
	{"PVMD",		0, 		           "PVMD_RUNS_HERE",0,0,TRUE, 0,0,0},
	{"W_STARTD",	"W_STARTD_LOG",   "WORLD_RUNS_HERE",0,0,TRUE, 0,0,0},
	{"W_SCHEDD",	"W_SCHEDD_LOG",   "WORLD_RUNS_HERE",0,0,TRUE, 0,0,0},
	};

Deamons::Deamons()
{
		no_deamons  =  sizeof(deamon_ptr)/ sizeof(deamon_ptr[0]);
}

void Deamons::InitParams()
{
	for ( int i =0; i < no_deamons; i++)
	{
		// pvm deamon shud be started only if it is specified in the
		// config file
		if ( deamon_ptr[i].flag_in_config_file != NULL )
		{
			char* tmp = param( deamon_ptr[i].flag_in_config_file);
			if ( tmp && (*tmp == 't' || *tmp == 'T'))
				deamon_ptr[i].flag = TRUE;
			else
			{
				deamon_ptr[i].flag = FALSE;
				continue;
			}
		}
			
		deamon_ptr[i].process_name = param(deamon_ptr[i].name_in_config_file);	
		if ( deamon_ptr[i].process_name == NULL )
		{
			dprintf(D_ALWAYS, "Process not found in config file: %s\n", 
					deamon_ptr[i].name_in_config_file);
			{ EXCEPT("Process name not found\n");}
		}

		// check that log file is necessary
		if ( deamon_ptr[i].log_filename_in_config_file != NULL)
		{
			deamon_ptr[i].log_name = 
					param(deamon_ptr[i].log_filename_in_config_file);
			if ( deamon_ptr[i].log_name == NULL )
			{
				dprintf(D_ALWAYS, "Process not found in config file: %s\n", 
						deamon_ptr[i].log_filename_in_config_file);
				{ EXCEPT("Log file name not found\n");}
			}
		}
	}
	int collect =  GetIndex("COLLECTOR");
	if ( collect == -1 )
	    	{ EXCEPT( "COLLECTOR not specified in config file" );}
    if ( collector_runs_here())
	    deamon_ptr[collect].flag = TRUE;
	else    deamon_ptr[collect].flag = FALSE;

	int negotiate = GetIndex("NEGOTIATOR");
	if ( negotiate == -1 )
			{ EXCEPT( "NEGOTIATOR not specified in config file" );}
	if ( negotiator_runs_here())
	    deamon_ptr[negotiate].flag = TRUE;
	else deamon_ptr[negotiate].flag = FALSE;

	// initialise master data structure
	int index = GetIndex("MASTER");
	if ( index == -1 )
	{
		dprintf(D_ALWAYS, "InitParams:MASTER not specified\n");
		EXCEPT("InitParams:MASTER not Specifed\n");
	}
	deamon_ptr[index].timeStamp = GetTimeStamp(deamon_ptr[index].process_name);
	deamon_ptr[index].pid = getpid();

	// if this is the world machine, then startd and startd are not run,
	// instead w-startd and w-schedd are run.
	index = GetIndex("W_SCHEDD");
	if ( index == -1 )
		dprintf(D_ALWAYS, " Know nothing about Condor Flock Configuration\n");
	else
	{
		// if WORLD_RUNS_HERE has been set to TRUE
		if ( deamon_ptr[index].flag == TRUE )
		{
			index = GetIndex("SCHEDD");
			if ( index != -1 ) deamon_ptr[index].flag = FALSE;
			index = GetIndex("STARTD");
			if ( index != -1 ) deamon_ptr[index].flag = FALSE;
			index = GetIndex("KBDD");
			if ( index != -1 ) deamon_ptr[index].flag = FALSE;
		}
	}
}

int Deamons::GetIndex(const char* name)
{
	for ( int i=0; i < no_deamons; i++)
		if ( strcmp(deamon_ptr[i].name_in_config_file,name) == 0 )
			return i;	
	return -1;
}

void Deamons::StartAllDeamons()
{
	// initialise master data structure
	int index = GetIndex("MASTER");
	if ( index == -1 )
	{
		dprintf(D_ALWAYS, "StartAllDeamons :MASTER not specified\n");
		EXCEPT("StartAllDeamons :MASTER not Specifed\n");
	}
	deamon_ptr[index].timeStamp = GetTimeStamp(deamon_ptr[index].process_name);
	deamon_ptr[index].pid       = getpid();

	for ( int i=0; i < no_deamons; i++)
	{
		// create only those processes that exists in this machine
		if ( deamon_ptr[i].flag == FALSE ) continue;
		deamon_ptr[i].pid = StartDeamon(deamon_ptr[i].process_name);
		deamon_ptr[i].timeStamp = GetTimeStamp(deamon_ptr[i].process_name);
	}
}

int Deamons::StartDeamon(char* process_name)
{
	int		pid;
	char	*shortname;

	if( NotFlag ) {
		dprintf( D_ALWAYS, "NOT Starting \"%s\"\n", process_name );
		return 0;
	}

	if( shortname = strrchr(process_name,'/') ) {
		shortname += 1;
	} else {
		shortname = process_name;
	}

	if( access(process_name,X_OK) != 0 ) {
		EXCEPT( "%s: Cannot execute", process_name );
	}

	if( (pid = vfork()) < 0 ) {
		EXCEPT( "vfork()" );
	}

	if( pid == 0 ) {	/* The child */
		/* Daemons need to be started as root. */
		set_root_euid();
		pid = getpid();
		if( setsid() == -1 ) {
			EXCEPT( "setsid(), errno = %d\n", errno );
		}
		(void)execl( process_name, shortname, "-f", 0 );
		/* Must be condor to write to log files. */
		set_condor_euid();
		EXCEPT( "execl( %s, %s, -f, 0 )", process_name, shortname );
		return 0;
	} else { 			/* The parent */
		dprintf( D_ALWAYS, "Started \"%s\", pid and pgroup = %d\n",
			shortname, pid );
		return pid;
	}
}

void Deamons::Restart(int pid)
{
	sleep(30);

	// find out which deamon died
	for ( int i=0; i < no_deamons; i++)
		if (( pid == deamon_ptr[i].pid)&&(deamon_ptr[i].flag == TRUE))
		{
			// this is the deamon that died: restart it
			dprintf( D_ALWAYS, "The %s (process %d ) died\n",
						deamon_ptr[i].name_in_config_file,pid);
			do_killpg( pid, SIGKILL ) ;
			if ( ++deamon_ptr[i].restarts > RestartsPerHour)
				give_up(deamon_ptr[i].process_name);

			// restart
			deamon_ptr[i].pid = StartDeamon(deamon_ptr[i].process_name);
			return;
		}
	 dprintf( D_ALWAYS, "Child %d died, but not a daemon -- Ignored\n", pid);
}


void Deamons::RestartMaster()
{
	int			pid;
	int			index = GetIndex("MASTER");

	if ( index == -1 )
	{
		dprintf(D_ALWAYS, "Restart Master:MASTER not specified\n");
		EXCEPT("Restart Master:MASTER not Specifed\n");
	}

	dprintf(D_ALWAYS, "RESTARTING MASTER (new executable)\n");
	install_sig_handler( SIGCHLD, (SIGNAL_HANDLER)SIG_IGN);

	for ( int i=0; i < no_deamons; i++)
		if ((deamon_ptr[i].flag == TRUE )&& ( deamon_ptr[i].pid )
				&& ( index != i))
		{
			do_killpg(deamon_ptr[i].pid, SIGKILL);
			dprintf(D_ALWAYS, "Killed %s pid = %d\n",
				deamon_ptr[i].name_in_config_file, deamon_ptr[i].pid);
		}

		/* Wait until all children die */
	for(;;) {
		pid = wait( 0 );
		dprintf( D_ALWAYS, "Wait() returns pid %d\n", pid );
		if( pid < 0 ) {
			if( errno == ECHILD ) {
				break;
			} else {
				EXCEPT( "wait( 0 )" );
			}
		}
	}
	dprintf( D_ALWAYS, "Done waiting for all children\n" );

	if ( MasterLock->release() == FALSE ) {
		dprintf(D_ALWAYS,"Cant remove lock on \"%s\"\n",deamon_ptr[index].log_name);
		EXCEPT( "file_lock(%d)", MasterLockFD );
	}
	dprintf( D_ALWAYS, "Unlocked file descriptor %d\n", MasterLockFD );
	(void)close( MasterLockFD );
	dprintf( D_ALWAYS, "Closed file descriptor %d\n", MasterLockFD );

	dprintf( D_ALWAYS, "Doing exec( \"%s\", \"condor_master\", 0 )", 
				deamon_ptr[index].process_name);
	(void)execl(deamon_ptr[index].process_name, "condor_master", 0);
	/* Must be condor to write to log files. */
	set_condor_euid();
	EXCEPT("execl(%s, condor_master, 0)",deamon_ptr[index].process_name);
}

void Deamons::SigalrmHandler() 
{

	dprintf( D_FULLDEBUG, "Entered sigalrm_handler()\n" );

	for  ( int j =0; j < no_deamons; j++)
	{
		int temp =  MAXIMUM(0,deamon_ptr[j].restarts);
		deamon_ptr[j].restarts = temp;
	}

	int master = GetIndex("MASTER");
	if ( master == -1 )
	{
		dprintf(D_ALWAYS, "SigalrmHandler:MASTER not specified\n");
		EXCEPT("SigalrmHandler:MASTER not Specifed\n");
	}
	if( NewExecutable(deamon_ptr[master].process_name,
									&deamon_ptr[master].timeStamp) ) 
	{
		dprintf(D_ALWAYS, " Restarting master\n");
		RestartMaster();
	}

	for (int i=0; i < no_deamons; i++)
	{
		if ( i== master ) continue;
		if (( deamon_ptr[i].flag == TRUE ) && ( deamon_ptr[i].pid )
				&& NewExecutable(deamon_ptr[i].process_name,
									&deamon_ptr[i].timeStamp))
		{
			dprintf(D_ALWAYS,"%s was modified.Killing %d\n", 
					deamon_ptr[i].name_in_config_file, deamon_ptr[i].pid);
			do_killpg( deamon_ptr[i].pid, SIGKILL );
			deamon_ptr[i].restarts = 0;
		}
	}
	dprintf( D_FULLDEBUG, "Calling event_mgr()\n" );
	event_mgr();

#if 0
		// do this every time for testing
	hourly_housekeeping();
#endif

}

char* Deamons::DeamonLog( int pid )
{
	// be careful : a pointer to data in this class is returned
	// posibility of getting tampered

	for ( int i=0; i < no_deamons; i++)
		if ( deamon_ptr[i].pid == pid )
			return (deamon_ptr[i].log_name);
	return "Unknown Program!!!";
}

char* Deamons::DeamonName( int pid )
{
	// be careful : a pointer to data in this class is returned
	// posibility of getting tampered

	for ( int i=0; i < no_deamons; i++)
		if ( deamon_ptr[i].pid == pid )
			return (deamon_ptr[i].process_name);
	return "Unknown Program!!!";
}
char* Deamons::SymbolicName( int pid )
{
	// be careful : a pointer to data in this class is returned
	// posibility of getting tampered

	for ( int i=0; i < no_deamons; i++)
		if ( deamon_ptr[i].pid == pid )
			return (deamon_ptr[i].name_in_config_file);
	return "Unknown Program!!!";
}

int Deamons::NumberRestarts(int pid)
{
	for ( int i=0; i < no_deamons; i++)
		if ( deamon_ptr[i].pid == pid )
			return (deamon_ptr[i].restarts);
	return -1;
}

void Deamons::SignalAll(int signal)
{
	// kills all deamons except the master itself
	for ( int i=0; i < no_deamons; i++)
		if ((deamon_ptr[i].flag == TRUE ) && ( deamon_ptr[i].pid )
		  && ( strcmp(deamon_ptr[i].name_in_config_file,"MASTER") != 0))
			do_killpg(deamon_ptr[i].pid, signal);
}
int Deamons::IsDeamon(int pid)
{
	for ( int i=0; i < no_deamons; i++)
		if ( deamon_ptr[i].pid == pid )
			return 1;
	return 0;
}
	
