/* 
** Copyright 1993 by Miron Livny, and Mike Litzkow
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:  Mike Litzkow
**
*/ 

#define _POSIX_SOURCE

#if 0
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include "types.h"
#include "util_lib_proto.h"
#include "proc_obj.h"
#include "filter.h"
#include "alloc.h"
#else
#include "condor_common.h"
#include "condor_debug.h"
#include "condor_jobqueue.h"
#include "condor_config.h"
#include "condor_updown.h"
#include "condor_disk.h"
#include "proc_obj.h"
#include "alloc.h"
#endif


static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */

DBM		*Q, *OpenJobQueue();

int		Long;				 // Use long display format
char	*MyName;			 // Name this program was called by
char	*Spool;				 // Name of job queue directory



	// Prototypes of local interest only
void GetJobStatus(char* user, int& running, int& idle, int& unexpanded);
void DisplayUserPriority(void);

void
GetJobStatus(char* user, int & running, int & idle, int & unexpanded)
{
	char	queue_name[_POSIX_PATH_MAX];
	char	*arg;
	List<ProcObj>	*proc_list;
	ProcObj			*p;
	ProcFilter *PFilter = new ProcFilter();

	running = idle = unexpanded = 0;

		/* Open job queue */
	(void)sprintf( queue_name, "%s/job_queue", Spool );
	if( (Q=OpenJobQueue(queue_name,O_RDONLY,0)) == NULL ) {
		EXCEPT( "OpenJobQueue(%s)", queue_name );
	}

	// DO NOT lock the job queue here.  If a user stops the process,
	// it will hang up all condor processes on the machine.


		// Create a list of all procs we're interested in
	PFilter->Install();
	proc_list = ProcObj::create_list( Q, ProcFilter::Func );

		// Display the list, and also count overall statistics
	proc_list->Rewind();
	while( p = proc_list->Next() ) 
		if ( strcmp(p->get_owner(), user) == 0 )
			switch( p->get_status() ) 
			{
			  case UNEXPANDED:
				unexpanded += 1;
				break;
			  case IDLE:
				idle += 1;
				break;
			  case RUNNING:
				running += 1;
				break;
			}

	ProcObj::delete_list( proc_list );
	delete PFilter;
	CloseJobQueue( Q );
}

void DisplayUserPriority(void)
{
	 char		filename[_POSIX_PATH_MAX];
	 UpDown updown;
	 User data;
	 int  running, idle, unexpanded;

	 // read updown priorities from disk file
	 sprintf(filename,"%s/%s",Spool, "UserPrio");
	 File ff(filename);
	 ff >> updown;		

	 // initialize list to be read from updown object
	 updown.StartUserInfo();

	 printf( " %-14s %-14s %-14s %-14s %-14s\n",
			"OWNER",
			"UPDOWN_PRIO",
			"RUNNING",
			"IDLE",
			"UNEXPANDED"
		);

	 // read all the users
	 while ( updown.GetNextUserInfo( data ) != Error )
	 {
		GetJobStatus(data.name, running, idle, unexpanded);
		printf(" %-14s %-14d %-14d %-14d %-14d\n",data.name,data.priority,
				running, idle, unexpanded);
		delete data.name;	// this is allocated in GetJobStatus
	}
}	

