/* 
** Copyright 1993 by Miron Livny, and Mike Litzkow
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the copyright holders not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the 
** copyright holders make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE COPYRIGHT HOLDERS DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE COPYRIGHT HOLDERS BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Author:  Mike Litzkow
**
*/ 

#define _POSIX_SOURCE

#if 0
#include <stdio.h>
#include <unistd.h>
#include <ctype.h>
#include <fcntl.h>
#include "types.h"
#include "util_lib_proto.h"
#include "proc_obj.h"
#include "filter.h"
#include "alloc.h"
#else
#include "condor_common.h"
#include "condor_debug.h"
#include "condor_jobqueue.h"
#include "condor_config.h"
#include "proc_obj.h"
#include "alloc.h"
#endif

static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */

DBM		*Q, *OpenJobQueue();

int		PrintSummary = TRUE; // Print a summary of jobs in the queue
int		Long;				 // Use long display format
char	*MyName;			 // Name this program was called by
char	*Spool;				 // Name of job queue directory

	// Summary information about the job queue
int		UnexpandedJobs;	
int		IdleJobs;
int		RunningJobs;

ProcFilter *PFilter = new ProcFilter();

	// Prototypes of local interest only
void print_summary();
void init_params();
void usage();
extern void DisplayUserPriority(void);    // display UPDOWN proirity

	// Tell folks how to use this program
void
usage()
{
	fprintf( stderr,
		"Usage: %s [-l] [cluster | cluster.proc | user_name ] ...\n", MyName
	);
	exit( 1 );
}


int
main( int argc, char *argv[] )
{
	char	queue_name[_POSIX_PATH_MAX];
	char	*arg;
	List<ProcObj>	*proc_list;
	ProcObj			*p;

	config( *argv, (CONTEXT *)0 );
	init_params();

	for( MyName = *argv++; arg = *argv; argv++ ) {
		if( arg[0] == '-' && arg[1] == 'l' ) {
			Long = TRUE;
			PrintSummary = FALSE;
		} else if ( arg[0] == '-' && arg[1] == 'u' )
		{
			DisplayUserPriority();			// UPDOWN  prio
			exit(0);
		} else {
			if( PFilter->Append(arg) ) {
				PrintSummary = FALSE;
			} else {
				usage();
			}
		}
	}


		/* Open job queue */
	(void)sprintf( queue_name, "%s/job_queue", Spool );
	if( (Q=OpenJobQueue(queue_name,O_RDONLY,0)) == NULL ) {
		EXCEPT( "OpenJobQueue(%s)", queue_name );
	}

	// DO NOT lock the job queue here.  If a user stops the process,
	// it will hang up all condor processes on the machine.


	if( !Long ) {
		printf( "%-7s %-14s %-11s %-12s %-2s %-3s %-4s %-18s\n\n",
		"   ID", " OWNER", "  SUBMITTED", "   CPU_USAGE",
		" ST", "PRI", "SIZE", "COMMAND" );
	}

		// Create a list of all procs we're interested in
	PFilter->Install();
	proc_list = ProcObj::create_list( Q, ProcFilter::Func );

		// Display the list, and also count overall statistics
	proc_list->Rewind();
	while( p = proc_list->Next() ) {
		if( Long ) {
			p->display();
		} else {
			p->display_short();
		}
		if( PrintSummary ) {
			switch( p->get_status() ) {
			  case UNEXPANDED:
				UnexpandedJobs += 1;
				break;
			  case IDLE:
				IdleJobs += 1;
				break;
			  case RUNNING:
				RunningJobs += 1;
				break;
			}
		}
	}

	ProcObj::delete_list( proc_list );
	delete PFilter;

	if( PrintSummary ) {
		print_summary();
	}

#if defined(ALLOC_DEBUG)
	print_alloc_stats();
#endif

	return 0;
}

extern "C" int
SetSyscalls( int foo ) { return foo; }




/*
  Find out where the job queue is stored on this machine by looking into
  configuration files.
*/
void
init_params()
{

	Spool = param("SPOOL");
	if( Spool == NULL ) {
		EXCEPT( "SPOOL not specified in config file\n" );
	}
}

/*
  Print a summary of all the jobs in the queue.
*/
void
print_summary()
{
	int		total;

	total = UnexpandedJobs + IdleJobs + RunningJobs;
	if( total == 0 ) {
		return;
	}

	printf( "\n%d %s in queue %d unexpanded, %d idle, %d running.\n",
	total, total>1 ? "jobs" : "job", UnexpandedJobs, IdleJobs, RunningJobs );
}
