/* 
** Copyright 1986, 1987, 1988, 1989, 1990, 1991 by the Condor Design Team
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the Condor Design Team not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the Condor
** Design Team make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE CONDOR DESIGN TEAM DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE CONDOR DESIGN TEAM BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Authors:  Allan Bricker and Michael J. Litzkow,
** 	         University of Wisconsin, Computer Sciences Dept.
** 
*/ 



#include <netdb.h>
#include <stdio.h>
#include <sys/types.h>
#include <utmp.h>
#include <nlist.h>
#include <signal.h>
#include <setjmp.h>

#include <pwd.h>
#include <errno.h>
#include <ctype.h>
#include <sys/stat.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/resource.h>

#if defined(AIX31) || defined(AIX32)
#	include <sys/statfs.h>
#elif defined(ULTRIX42) || defined(ULTRIX43)
#	include <sys/mount.h>
#elif defined(IRIX331)
#	include <sys/statfs.h>
#elif defined(OSF1)
	/* nothing */
#else
#	include <sys/vfs.h>
#endif

#if defined(HPUX9)
#include <unistd.h>
#include <sys/utsname.h>
#endif

#include <sys/file.h>
#include <netinet/in.h>
#include <rpc/types.h>
#include <rpc/xdr.h>

#ifdef DYNIX
#include <sys/tmp_ctl.h>
#include <sys/vm.h>
#endif DYNIX


#include <strings.h>
#include <arpa/inet.h>
#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 100
#endif MAXHOSTNAMELEN

#include "sched.h"
#include "debug.h"
#include "trace.h"
#include "except.h"
#include "expr.h"
#include "clib.h"

#define STANDARD 1
#define VANILLA 5

static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */

CONTEXT		*create_context();
bool_t 		xdr_int();
XDR			*xdr_Udp_Init(), *xdr_Init();
ELEM		*create_elem();
EXPR		*create_expr(), *build_expr();;
float		calc_load_avg();
char		*strdup();

char		*param();
char *		get_full_hostname();

typedef void (*SIG_HANDLER)();
void install_sig_handler( int sig, SIG_HANDLER handler );


extern CONTEXT	*MachineContext;
extern CONTEXT	*GenericJob;
extern CONTEXT	*JobContext;
extern int		StarterPid;
extern int		LastTimeout;
extern char		*MgrHost;
extern char		*CondorAdministrator;
extern int		UdpSock;
extern int      UdpSockAlternate;
extern int		Memory;
extern char		*Execute;
extern char		*Starter;
extern char		*KbdDev;
extern char		*MouseDev;
extern int		State;
extern int		PollsPerUpdate;
extern char*	AlternateCollectorHost;
extern int		JobUniverse;
static int		Polls = 0;

char			*ClientMachine;
char			*Mail;

int		Last_X_Event;

job_dies()
{
	StarterPid = 0;
	if( JobContext ) {
		free_context( JobContext );
		JobContext = NULL;
	}

#define TESTING FALSE
#if !TESTING
	cleanup_execute_dir();
#endif

	FREE( ClientMachine );
	ClientMachine = NULL;
	change_states( NO_JOB );
	Polls = 0;

	timeout();
}

timeout()
{



	evaluate_situation();

	if( Polls == 0 ) {
		(void)update_central_mgr();
	}
	if( ++Polls >= PollsPerUpdate ) {
		Polls = 0;
	}
	LastTimeout = (int)time( (time_t *)0 );

}


#define ABORT \
	if( job_context ) {				\
		free_context( job_context );\
	}								\
	if( ClientMachine ) {			\
		FREE( ClientMachine );		\
		ClientMachine = NULL;		\
	}								\
	return;

start_job_request( xdrs, from )
XDR		*xdrs;
struct sockaddr_in	*from;
{
	int		start, job_reqs;
	CONTEXT	*job_context = NULL;
	struct hostent  *hp, *gethostbyaddr();
    ELEM    tmp;

    if( (hp=gethostbyaddr((char *)&from->sin_addr,
					sizeof(struct in_addr), from->sin_family)) == NULL ) {
        EXCEPT(  "Can't find host name" );
    }
	ClientMachine = strdup( hp->h_name );


	dprintf( D_ALWAYS, "Got start_job_request from %s\n", ClientMachine );

		/* Make sure we have up to date swap space information */
	tmp.type = INT;
	tmp.i_val = calc_virt_memory();
	store_stmt( build_expr("VirtualMemory",&tmp), MachineContext );

		/* Read in job facts and booleans */
	job_context = create_context();
	if( !xdr_context(xdrs,job_context) ) {
		dprintf( D_ALWAYS, "Can't receive context from shadow\n" );
		return;
	}
	if( !xdrrec_skiprecord(xdrs) ) {
		dprintf( D_ALWAYS, "Can't receive context from shadow\n" );
		return;
	}

	dprintf( D_JOB, "JOB_CONTEXT:\n" );
	if( DebugFlags & D_JOB ) {
		display_context( job_context );
	}
	dprintf( D_JOB, "\n" );

	dprintf( D_MACHINE, "MACHINE_CONTEXT:\n" );
	if( DebugFlags & D_MACHINE ) {
		display_context( MachineContext );
	}
	dprintf( D_MACHINE, "\n" );

	if( State != NO_JOB ) {
		dprintf( D_ALWAYS, "State != NO_JOB\n" );
		(void)reply( xdrs, NOT_OK );
		ABORT;
	}

		/* See if machine and job meet each other's requirements, if so
		   start the job and tell shadow, otherwise refuse and clean up */
	if( evaluate_bool("START",&start,MachineContext,job_context) < 0 ) {
		start = 0;
	}
	if( evaluate_bool("JOB_REQUIREMENTS",&job_reqs,MachineContext,job_context)
																		< 0 ) {
		job_reqs = 0;
	}

	if( !start || !job_reqs ) {
		dprintf( D_ALWAYS, "start = %d, job_reqs = %d\n", start, job_reqs );
		(void)reply( xdrs, NOT_OK );
		ABORT;
	}

		/* We could have been asked to run an alternate version of the
		   starter which is not listed in our config file.  If so, it
		   would show up as NULL here */
	if( Starter == NULL ) {
		dprintf( D_ALWAYS, "Alternate starter not found in config file\n" );
		(void)reply( xdrs, NOT_OK );
		ABORT;
	}
		

	if( !reply(xdrs,OK) ) {
		ABORT;
	}
	JobContext = job_context;
	if( start_job(xdrs) < 0 ) {
		ABORT;
	}
	change_states( JOB_RUNNING );
	Polls = 0;

            /* PREEMPTION : dhruba */
    insert_context("ClientMachine", ClientMachine, MachineContext);
	tmp.type = INT;
	tmp.i_val = (int)time( (time_t *)0 );
	store_stmt( build_expr("JobStart",&tmp), MachineContext );

	timeout();
}
#undef ABORT

x_event( xdrs, from )
XDR		*xdrs;
struct sockaddr_in	*from;
{
	Last_X_Event = time( (time_t *)0 );
	timeout();
}


evaluate_situation()
{
	int		tmp;
	int		start = FALSE;
	int		block_requested;

	evaluate_load();
	block_requested = condor_block_requested();

	/*
	dprintf( D_ALWAYS, "----------------\n" );
	display_context( MachineContext );
	dprintf( D_ALWAYS, "----------------\n" );
	display_context( GenericJob );
	dprintf( D_ALWAYS, "----------------\n" );
	*/

		/* We have a blocker, but not holding a job */
	if( State == SYSTEM ) {
		if( !block_requested ) {
			change_states( NO_JOB );
		}
	}

		/* Not holding a job, no blockers */
	if( State == NO_JOB ) {
		if( block_requested ) {
			change_states( SYSTEM );
			return;
		}
		(void)evaluate_bool( "START", &start, MachineContext, GenericJob );
		if( start ) {
			set_machine_status( M_IDLE );
			dprintf( D_FULLDEBUG, "Set machine status to M_IDLE\n" );
		} else {
			set_machine_status( USER_BUSY );
			dprintf( D_FULLDEBUG, "Set machine status to USER_BUSY\n" );
		}
	}

	if( StarterPid && !JobContext ) {
		EXCEPT( "evaluate_situation, StarterPid = %d, but JobContext is NULL\n",
					StarterPid );
	}

	if( !JobContext ) {
		return;
	}

	if( State == CHECKPOINTING ) {
		if ( JobUniverse == VANILLA ) {
			if( evaluate_bool("KILL_VANILLA",&tmp,MachineContext,JobContext) < 0 ) {
				if( evaluate_bool("KILL",&tmp,MachineContext,JobContext) < 0 ) {
					EXCEPT( "Can't evaluate machine boolean \"KILL\"\n" );
				}
			}
		} else {
			if( evaluate_bool("KILL",&tmp,MachineContext,JobContext) < 0 ) {
				EXCEPT( "Can't evaluate machine boolean \"KILL\"\n" );
			}
		}
		if( tmp ) {
			kill_job( NOT_OK );
			return;
		}
	}

	if( State == SUSPENDED ) {
		if ( JobUniverse == VANILLA ) {
			if( evaluate_bool("VACATE_VANILLA",&tmp,MachineContext,JobContext) < 0 ) {
				if( evaluate_bool("VACATE",&tmp,MachineContext,JobContext) < 0 ) {
					EXCEPT( "Can't evaluate machine boolean \"VACATE\"\n" );
				}
			}
		} else {
			if( evaluate_bool("VACATE",&tmp,MachineContext,JobContext) < 0 ) {
				EXCEPT( "Can't evaluate machine boolean \"VACATE\"\n" );
			}
		}
		if( tmp ) {
			vacate_order();
			return;
		}
	}

	if( State == JOB_RUNNING ) {
		if( block_requested ) {
			block_order();
			return;
		}
		if ( JobUniverse == VANILLA ) {
			if( evaluate_bool("SUSPEND_VANILLA",&tmp,MachineContext,JobContext) < 0 ) {
				if( evaluate_bool("SUSPEND",&tmp,MachineContext,JobContext) < 0 ) {
					EXCEPT( "Can't evaluate machine boolean \"SUSPEND\"\n" );
				}
			}
		} else {
			if( evaluate_bool("SUSPEND",&tmp,MachineContext,JobContext) < 0 ) {
				EXCEPT( "Can't evaluate machine boolean \"SUSPEND\"\n" );
			}
		}
		if( tmp ) {
			suspend_order();
			return;
		}
	}

		/* Holding a job, and blocked */
	if( State == BLOCKED ) {
		if( !block_requested ) {
			unblock_order();
			return;
		}
	}

	if( State == SUSPENDED ) {
		if ( JobUniverse == VANILLA ) {
			if( evaluate_bool("CONTINUE_VANILLA",&tmp,MachineContext,JobContext) < 0 ) {
				if( evaluate_bool("CONTINUE",&tmp,MachineContext,JobContext) < 0 ) {
					EXCEPT( "Can't evaluate machine boolean \"CONTINUE\"\n" );
				}
			}
		} else {
			if( evaluate_bool("CONTINUE",&tmp,MachineContext,JobContext) < 0 ) {
				EXCEPT( "Can't evaluate machine boolean \"CONTINUE\"\n" );
			}
		}
		if( tmp ) {
			resume_order();
			return;
		}
	}

}

update_central_mgr()
{
	int		cmd;
	XDR		xdr, *xdrs;
    ELEM    tmp;
	static int	first_time = 1;
	char	*calc_subnet_name();
	char	*max_jobs;
	int		phys_memory;
	int     now_idle;

#if defined(BENCHMARK)
	static int calc_mips = 0;
	static int calc_kflops = 0;
	static int consecutive_idle = 0;
	int		new_mips_calc;
	int     new_kflops_calc;
#endif

#if defined(HPUX9)
	struct utsname utsinfo;
#endif


	if( first_time ) {

		/* Probably have to be root to read /dev/kmem */
		set_root_euid(__FILE__,__LINE__);
			phys_memory = calc_phys_memory();
		set_condor_euid(__FILE__,__LINE__);

		if( phys_memory > 0 ) {
			tmp.type = INT;
			tmp.i_val = phys_memory;
			store_stmt( build_expr("Memory",&tmp), MachineContext );
		}
	}


	now_idle = machine_idle();

		/* Calculating virt memory is expensive, only do it if we are idle */
	if( now_idle || first_time ) {
		tmp.type = INT;
		tmp.i_val = calc_virt_memory();
		store_stmt( build_expr("VirtualMemory",&tmp), MachineContext );
#if !defined(BENCHMARK)
		first_time = 0;
#endif

	}

#if defined(BENCHMARK)
	if ( now_idle ) {
		consecutive_idle++;
	} else {
		consecutive_idle = 0;
	}


	/* Calculating speed is even more expensive, do it only when we've been
	   idle for a while */
	if ( consecutive_idle >= 5 || first_time) {
		first_time = 0;
		consecutive_idle = 0;
		new_mips_calc = dhry_mips();
		new_kflops_calc = clinpack_kflops();
		if ( calc_mips == 0) {
			calc_mips = new_mips_calc;
			calc_kflops = new_kflops_calc;
		} else {
			calc_mips = (calc_mips * 3 + new_mips_calc) / 4;
			calc_kflops = (calc_kflops * 3 + new_kflops_calc) / 4;
		}

		dprintf( D_ALWAYS, "re-calc:DHRY_MIPS = %d, CLINPACK KFLOPS = %d\n",
				new_mips_calc, new_kflops_calc );
		tmp.type = INT;
		tmp.i_val = calc_mips;
		store_stmt( build_expr("MIPS", &tmp), MachineContext );
		tmp.type = INT;
		tmp.i_val = calc_kflops;
		store_stmt( build_expr("KFLOPS", &tmp), MachineContext );
		first_time = 0;
	}
#endif

	tmp.type = STRING;
	tmp.s_val = calc_subnet_name();
	if( strlen(tmp.s_val) != 0 )
		store_stmt( build_expr("Subnet", &tmp), MachineContext );
	free( tmp.s_val );

	tmp.type = STRING;
	tmp.s_val = (char *)param("FLAVOR");		/* rwong */
#if defined(HPUX9)
    /* for HPs, we will get the "flavor", or model type, from the kernel */
    /* however,if the user actually specifies a flavor,believe the human!*/
    if ( tmp.s_val == NULL ) {
        if ( uname((struct utsname *) &utsinfo) != -1 )
            tmp.s_val = utsinfo.machine;
    }
#endif
	if( tmp.s_val )
		store_stmt( build_expr("Flavor", &tmp), MachineContext );


	tmp.type = INT;

	max_jobs = (char *)param( "MAX_JOBS_RUNNING" );		/* rwong */
	if( max_jobs ) {
		tmp.i_val = atoi( max_jobs );
		free( max_jobs );
	} else {
		tmp.i_val = 15;
	}

	store_stmt( build_expr("MAX_JOBS_RUNNING", &tmp), MachineContext);

	xdrs = xdr_Udp_Init( &UdpSock, &xdr );

	xdrs->x_op = XDR_ENCODE;


		/* Send the command */
	dprintf( D_XDR, "About to send update to collector\n" );
	cmd = STARTD_INFO;


	if( !xdr_int(xdrs, &cmd) ) {
		xdr_destroy( xdrs );
		return -1;
	}


	if( !xdr_context(xdrs,MachineContext) ) {
		xdr_destroy( xdrs );
		return -1;
	}

	if( !xdrrec_endofrecord(xdrs,TRUE) ) {
		xdr_destroy( xdrs );
		return -1;
	}

	if ( AlternateCollectorHost != NULL )
		updateAlternateCollector(MachineContext);
	xdr_destroy( xdrs );
	dprintf( D_XDR, "Done sending update to collector\n" );

	dprintf(D_FULLDEBUG, " UpdateCentral Manager\n");
	if ( DebugFlags & D_PREEMPT ) display_context(MachineContext);
	return 0;
}

change_states( new_state )
int	new_state;
{
	ELEM	tmp;
	char	*name;
	int		start = FALSE;

	switch( new_state ) {
		case NO_JOB:
			name = "NoJob";
			(void)evaluate_bool( "START", &start, MachineContext, GenericJob );
			if( start ) {
				set_machine_status( M_IDLE );
			} else {
				set_machine_status( USER_BUSY );
			}
			/* PREEMPTION : dhruba */
    		insert_context("ClientMachine", "", MachineContext);

			break;
		case JOB_RUNNING:
			name = "Running";
			set_machine_status( JOB_RUNNING );
			break;
		case KILLED:
			name = "Killed";
			set_machine_status( KILLED );
			break;
		case CHECKPOINTING:
			name = "Checkpointing";
			set_machine_status( CHECKPOINTING );
			break;
		case SUSPENDED:
			name = "Suspended";
			set_machine_status( SUSPENDED );
			break;
		case BLOCKED:
			name = "Blocked";
			set_machine_status( BLOCKED );
			break;
		case SYSTEM:
			name = "System";
			set_machine_status( SYSTEM );
			break;
		default:
			EXCEPT( "Change states, unknown state (%d)", new_state );
	}

	State = new_state;

	tmp.type = STRING;
	tmp.s_val = name;
	store_stmt( build_expr("State",&tmp), MachineContext );

	tmp.type = INT;
	tmp.i_val = (int)time( (time_t *)0 );
	store_stmt( build_expr("EnteredCurrentState",&tmp), MachineContext );
}

kill_job( status )
int		status;
{
	if( status == OK ) {
		dprintf( D_ALWAYS, "Called kill_job( OK )\n" );
	} else {
		dprintf( D_ALWAYS, "Called kill_job( NOT_OK )\n" );
	}
	if( !StarterPid ) {
		dprintf( D_ALWAYS, "Called kill_job, but no starter process active\n" );
		return;
	}

#define MAIL_STUCK_JOB_MESSAGE
#ifdef MAIL_STUCK_JOB_MESSAGE
	if( status != OK ) {
		mail_stuck_job_message();
	}
#endif MAIL_STUCK_JOB_MESSAGE

		/* Could be the starter will die naturally just before we try to
		   send the sig, so have to ignore errors here... */
	if( status == OK ) {
		(void) kill_starter( StarterPid, SIGINT );
		change_states( CHECKPOINTING );
	} else {
		dprintf( D_ALWAYS, "Killing myself and my whole process group...\n" );
		set_root_euid();
#if defined(HPUX9)
		if( kill(-getpgrp(),SIGKILL) < 0 ) {
#else
		if( kill(-getpgrp(0),SIGKILL) < 0 ) {
#endif 
			EXCEPT( "kill( my_process_group ) euid = %d", geteuid() );
		}
	
		sigpause( 0 );
	}
}

/*
  We have been hosting a job, decided it should leave and sent it the
  appropriate signal, but it refuses to go.  Now we try to mail some human
  so they know about the problem.  If anything goes wrong in here, don't
  do any EXCEPTS.  The routine which calls us wants to take the emergency
  step of killing ourself and our whole process group.  An error here
  and an EXCEPT could leave the job running amok.
*/
mail_stuck_job_message()
{
	char	cmd[512];
	char	hostname[512];
	FILE	*mailer, *popen();

	dprintf( D_ALWAYS, "Mailing administrator (%s)\n", CondorAdministrator );

	if( gethostname(hostname,sizeof(hostname)) < 0 ) {
		dprintf( D_ALWAYS, "gethostname failed\n" );
		return;
	}

	(void)sprintf( cmd, "%s %s", Mail, CondorAdministrator );
	if( (mailer=popen(cmd,"w")) == NULL ) {
		dprintf( D_ALWAYS, "popen \"%s\" failed\n", cmd );
		return;
	}

	fprintf( mailer, "To: %s\n", CondorAdministrator );
	fprintf( mailer, "Subject: Condor Problem\n" );
	fprintf( mailer, "\n" );

	fprintf( mailer, "The startd on \"%s\" is killing a Condor job ",
				hostname );
	fprintf( mailer, "from \"%s\"\n", ClientMachine );
	fprintf( mailer, "because it didn't clean up after being sent a TSTP.\n"  );
	fprintf( mailer, "\n" );
	mail_context( "JobContext:", JobContext, mailer );
	fprintf( mailer, "\n" );
	mail_context( "MachineContext:", MachineContext, mailer );
	fprintf( mailer, "\n" );

	fprintf( mailer, "CurrentTime = %d\n", (int)time((time_t *)0) );

		/* don't use 'pclose()' here, it does its own wait, and messes
	       with our handling of SIGCHLD! */
		/* except on HPUX pclose() is both safe & required - Todd 12/94 */
#if defined(HPUX9)
	pclose( mailer );
#else
	(void)fclose( mailer );
#endif

		/* The routine we are called by is going to kill our whole process
		   group now.  Give the mailer time to do its job first... */
	sleep( 90 );
}

ckpt_order()
{
	dprintf( D_ALWAYS, "Called ckpt_order()\n" );

	if( !StarterPid ) {
		dprintf( D_ALWAYS,
					"Called ckpt_order, but no starter process active\n" );
	} else {
			/* Could be the starter will die naturally just before we try to
			   send the sig, so have to ignore errors here... */
		(void) kill_starter( StarterPid, SIGUSR2 );
		change_states( CHECKPOINTING );
	}
}

vacate_order()
{
	dprintf( D_ALWAYS, "Called vacate_order()\n" );

	dprintf( D_MACHINE, "MACHINE_CONTEXT:\n" );
	if( DebugFlags & D_MACHINE ) {
		display_context( MachineContext );
	}
	dprintf( D_MACHINE, "\n" );

	if( !StarterPid ) {
		dprintf( D_ALWAYS,
					"Called vacate_order, but no starter process active\n" );
	} else {
			/* Could be the starter will die naturally just before we try to
			   send the sig, so have to ignore errors here... */
		(void) kill_starter( StarterPid, SIGTSTP );
		change_states( CHECKPOINTING );
	}
}

block_order()
{
	dprintf( D_ALWAYS, "Called block_order()\n" );

	if( !StarterPid ) {
		dprintf( D_ALWAYS,
					"Called block, but no starter process active\n" );
	} else {
			/* Could be the starter will die naturally just before we try to
			   send the sig, so have to ignore errors here... */
		(void) kill_starter( StarterPid, SIGUSR1 );
		change_states( BLOCKED );
	}
}

unblock_order()
{
	dprintf( D_ALWAYS, "Called unblock_order()\n" );

	change_states( SUSPENDED );
}

suspend_order()
{
	dprintf( D_ALWAYS, "Called suspend_order()\n" );

	dprintf( D_MACHINE, "MACHINE_CONTEXT:\n" );
	if( DebugFlags & D_MACHINE ) {
		display_context( MachineContext );
	}
	dprintf( D_MACHINE, "\n" );

	if( !StarterPid ) {
		dprintf( D_ALWAYS,
					"Called suspend_order, but no starter process active\n" );
	} else {
			/* Could be the starter will die naturally just before we try to
			   send the sig, so have to ignore errors here... */
		(void) kill_starter( StarterPid, SIGUSR1 );
		change_states( SUSPENDED );
	}
}

resume_order()
{
	dprintf( D_ALWAYS, "Called resume_order()\n" );

	dprintf( D_MACHINE, "MACHINE_CONTEXT:\n" );
	if( DebugFlags & D_MACHINE ) {
		display_context( MachineContext );
	}
	dprintf( D_MACHINE, "\n" );

	if( !StarterPid ) {
		dprintf( D_ALWAYS,
					"Called resume_order, but no starter process active\n" );
	} else {
			/* Could be the starter will die naturally just before we try to
			   send the sig, so have to ignore errors here... */
		(void) kill_starter( StarterPid, SIGCONT );
		change_states( JOB_RUNNING );
	}
}


reply( xdrs, answer )
XDR	*xdrs;
int	answer;
{
	xdrs->x_op = XDR_ENCODE;
	if( !xdr_int(xdrs,&answer) ) {
		return FALSE;
	}
	if( !xdrrec_endofrecord(xdrs,TRUE) ) {
		return FALSE;
	}
	dprintf( D_ALWAYS, "Replied %s\n", answer ? "ACCEPTED" : "REFUSED" );
	return TRUE;
}

evaluate_load()
{
    ELEM    tmp;



    tmp.type = INT;
    tmp.i_val = calc_memory();

    store_stmt( build_expr("Memory",&tmp), MachineContext );

    tmp.i_val = calc_disk();
    store_stmt( build_expr("Disk",&tmp), MachineContext );

    tmp.i_val = user_idle_time();
    store_stmt( build_expr("KeyboardIdle",&tmp), MachineContext );

    tmp.i_val = calc_Ncpus();
    store_stmt( build_expr("Cpus",&tmp), MachineContext );

	tmp.type = FLOAT;
    tmp.f_val = calc_load_avg();
    store_stmt( build_expr("LoadAvg",&tmp), MachineContext );

	/*
	** display_context( MachineContext );
	*/
}

/*
** Calculate the main memory on our machine in megabytes.  For now we'll
** just look it up in the config file, someday we should probe the kernel
** for this information.
*/
calc_memory()
{
    return Memory;
}

/*
** Calculate the free disk in the filesystem which we use for hosting
** foreign executables in kilobytes.  We don't give the real answer,
** instead we allow a little "slop".
*/
#define FS_SLOP 512     /* Kilobytes */
calc_disk()
{
	return free_fs_blocks(Execute) - FS_SLOP;
}


#if defined(OSF1)
#	define MAXINT ((1<<30)-1)
#else
#	define MAXINT ((1<<31)-1)
#endif

#ifndef MIN
#define MIN(a,b) ((a)<(b)?(a):(b))
#endif

#if defined(OSF1)
	static char *UtmpName = "/var/adm/utmp";
#else
	static char *UtmpName = "/etc/utmp";
#endif


user_idle_time()
{
    int     tty_idle;
    int     user_idle;
	int		now;
    struct utmp utmp;
#if defined(HPUX9)
    int i;
    char devname[MAXPATHLEN];
#endif

	user_idle = tty_pty_idle_time();
	dprintf( D_FULLDEBUG, "ttys/ptys idle %d seconds\n",  user_idle );

	if( KbdDev ) {
        tty_idle = tty_idle_time( KbdDev );
        user_idle = MIN( tty_idle, user_idle );
		dprintf( D_FULLDEBUG, "/dev/%s idle %d seconds\n", KbdDev, tty_idle );
	}

	if( MouseDev ) {
        tty_idle = tty_idle_time( MouseDev );
        user_idle = MIN( tty_idle, user_idle );
		dprintf( D_FULLDEBUG, "/dev/%s idle %d seconds\n", MouseDev, tty_idle );
	}
/*
** On HP's "/dev/hil1" - "/dev/hil7" are mapped to input devices such as
** a keyboard, mouse, touch screen, digitizer, etc.We can then conveniently
** stat those, and avoid having to run the condor_kbd at all.
*/
#if defined(HPUX9)
        for( i=1; i<=7; i++ ) {
                sprintf( devname, "hil%d", i );
                tty_idle = tty_idle_time( devname );
                user_idle = MIN( tty_idle, user_idle );
                dprintf( D_FULLDEBUG, "%s idle %d seconds\n", devname, tty_idle
);
        }
#endif


    now = (int)time( (time_t *)0 );
	user_idle = MIN( now - Last_X_Event, user_idle );

	dprintf( D_FULLDEBUG, "User Idle Time is %d seconds\n", user_idle );
    return user_idle;
}

#if !defined(OSF1) /* in tty_idle.OSF1.C for that platform */
#include <limits.h>
tty_pty_idle_time()
{
    FILE    *fp;
    int     tty_idle;
    int     answer    = INT_MAX;
	int		now;
    struct utmp utmp;

    if( (fp=fopen(UtmpName,"r")) == NULL ) {
        EXCEPT( "fopen of \"%s\"", UtmpName );
    }

    while( fread( (char *)&utmp, sizeof utmp, 1, fp ) ) {
#if defined(AIX31) || defined(AIX32) || defined(IRIX331)
		if( utmp.ut_type != USER_PROCESS )
#else
		if( utmp.ut_name[0] == '\0' )
#endif
            continue;

        tty_idle = tty_idle_time( utmp.ut_line );
        answer = MIN( tty_idle, answer );
    }
    (void)fclose( fp );

	return answer;
}
#endif

calc_Ncpus()
{
#ifdef sequent
    int     cpus = 0;
    int     eng;

    if( (eng=tmp_ctl(TMP_NENG,0)) < 0 ) {
        perror( "tmp_ctl(TMP_NENG,0)" );
        exit( 1 );
    }

    while( eng-- ) {
        if( tmp_ctl(TMP_QUERY,eng) == TMP_ENG_ONLINE ) {
            cpus++;
        }
    }
    return cpus;
#else sequent
    return 1;
#endif sequent
}


tty_idle_time( file )
char    *file;
{
    struct stat buf;
    long        now;
    long        answer;
    static char     pathname[100] = "/dev/";

    (void)strcpy( &pathname[5], file );
    if( stat(pathname,&buf) < 0 ) {
		dprintf( D_FULLDEBUG, "Error on stat(%s,0x%x), errno = %d\n",
			pathname, &buf, errno );
		buf.st_atime = 0;
    }

    now = (int)time( (time_t *)0 );
    answer = now - buf.st_atime;

		/* This only happens if somebody screws up the date on the
		   machine :-<  */
	if( answer < 0 ) {
		answer = 0;
	}

    /*
    printf( "%s: %d seconds idle\n", file, answer );
    */
    return answer;
}

mail_context( name, context, fp )
char	*name;
CONTEXT	*context;
FILE	*fp;
{
    int     i, j;
	EXPR	*expr;

	fprintf( fp, "%s\n", name );
    for( i=0; i<context->len; i++ ) {
        fprintf( fp, "Stmt %2d:", i );
		expr = context->data[i];

        for( j=0; j<expr->len; j++ ) {
            display_elem( expr->data[j], fp );
            if( j+1 < expr->len ) {
                (void)fputc( ' ', fp );
            }
        }
        (void)fputc( '\n', fp );
    }
}

extern int	errno;
kill_starter( pid, signo )
int pid, signo;
{
	struct stat buf;

	for( errno=0; stat(Starter, &buf) < 0; errno=0 ) {
		if( errno == ETIMEDOUT ) {
			continue;
		}
		EXCEPT( "kill_starter( %d, %d ): cannot stat <%s> -- errno = %d\n",
					pid, signo, Starter, errno );
	}

#ifdef NFSFIX
	/* Must be root to kill starter.  Starter will have client uid and gid. */
	set_root_euid(__FILE__,__LINE__);
#endif NFSFIX
		/* send continue first in case process is stopped */
	if( signo != SIGSTOP && signo != SIGCONT ) {
		(void) kill( pid, SIGCONT );
	}
	(void) kill( pid, signo );
#ifdef NFSFIX
	/* Back to being condor. */
	set_condor_euid(__FILE__,__LINE__);
#endif NFSFIX
}

my_abort()
{
	install_sig_handler( SIGQUIT, SIG_DFL );
	(void)kill( getpid(), SIGQUIT );
	(void)sigpause( 0 );
}

char *
calc_subnet_name()
{
	struct hostent	*hostp;
	char			hostname[MAXHOSTNAMELEN];
	char			subnetname[MAXHOSTNAMELEN];
	char			*subnet_ptr;
	char			*host_addr_string;
	int				subnet_length;
	struct in_addr	in;

	if( gethostname(hostname, sizeof(hostname)) == -1 ) {
		dprintf( D_ALWAYS, "Gethostname failed");
		return strdup("");
	}
	if( (hostp = gethostbyname(hostname)) == NULL ) {
		dprintf( D_ALWAYS, "Gethostbyname failed");
		return strdup("");
	}
	bcopy((char *)hostp->h_addr, (char *) &in, hostp->h_length);
	host_addr_string = inet_ntoa( in );
	if( host_addr_string ) {
		subnet_ptr = (char *) rindex(host_addr_string, '.');
		if(subnet_ptr == NULL) {
			return strdup("");
		}
		subnet_length = subnet_ptr - host_addr_string;
		strncpy(subnetname, host_addr_string, subnet_length);
		subnetname[subnet_length] = '\0';
		return (strdup(subnetname));
	}
	return strdup("");
}

machine_idle()
{
	int		start;
	char	*state;

	if( evaluate_bool("START",&start,MachineContext,GenericJob) < 0 ) {
		EXCEPT( "Can't evaluate START expression against GenericJob" );
	}
	if( !start ) {
		return 0;
	}

	if( evaluate_string("State",&state,MachineContext,GenericJob) < 0 ) {
		EXCEPT( "Can't evaluate State variable" );
	}
		
	return( strcmp("NoJob",state) == 0 );
}

new_proc_order()
{
	dprintf( D_ALWAYS, "Called new_proc_order()\n" );

	if( !StarterPid ) {
		dprintf( D_ALWAYS,
			"Called new_proc_order, but no starter process active\n"
		);
	} else {
		(void) kill_starter( StarterPid, SIGHUP );
	}
}

/*
** Sometimes, we require that the startd send info to an alternate
** collector over and above the regular collector. This helps in
** gathering statistics
*/
int
updateAlternateCollector( context)
CONTEXT*    context;
{
    int     cmd;
    XDR     xdr, *xdrs;
    xdrs = xdr_Udp_Init( &UdpSockAlternate, &xdr );
    xdrs->x_op = XDR_ENCODE;
    dprintf( D_XDR, "About to send update to alternate collector\n" );

    cmd = STARTD_INFO;

    if( !xdr_int(xdrs, &cmd) ) {
        xdr_destroy( xdrs );
        return -1;
    }


    if( !xdr_context(xdrs,context) ) {
        xdr_destroy( xdrs );
        return -1;
    }

    if( !xdrrec_endofrecord(xdrs,TRUE) ) {
        xdr_destroy( xdrs );
        return -1;
    }

    xdr_destroy( xdrs );
    dprintf( D_XDR, "Done sending update to alternate collector\n" );

    return 0;
}

char *get_host_cell();

init_fs_info()
{
	ELEM	tmp;
	char	*ptr;
	char	*host_cell;

	host_cell = get_host_cell();

	tmp.type = STRING;
	tmp.s_val = host_cell;
	if( tmp.s_val ) {
		store_stmt( build_expr("AFS_Cell",&tmp), MachineContext );
		dprintf( D_ALWAYS, "AFS_Cell = \"%s\"\n", tmp.s_val );
	} else {
		dprintf( D_ALWAYS, "AFS_Cell not set\n" );
	}

	/*
	  If the uid domain is not defined, accept uid's only from this
	  machine.
	*/
	if( (ptr = param("UID_DOMAIN")) == NULL ) {
		ptr = get_full_hostname();
	}

	/*
	  If the uid domain is defined as "*", accept uid's from anybody.
	*/
	if( ptr[0] == '*' ) {
		ptr = "";
	}

	tmp.type = STRING;
	tmp.s_val = ptr;
	store_stmt( build_expr("UidDomain",&tmp), MachineContext );
	dprintf( D_ALWAYS, "UidDomain = \"%s\"\n", tmp.s_val );

	/*
	  If the filesystem domain is not defined, we share files only
	  with our own machine.
	*/
	if( (ptr = param("FILESYSTEM_DOMAIN")) == NULL ) {
		ptr = get_full_hostname();
	}

	/*
	  If the filesystem domain is defined as "*", assume we share
	  files with everybody.
	*/
	if( ptr[0] == '*' ) {
		ptr = "";
	}

	tmp.type = STRING;
	tmp.s_val = ptr;
	store_stmt( build_expr("FilesystemDomain",&tmp), MachineContext );
	dprintf( D_ALWAYS, "FilesystemDomain = \"%s\"\n", tmp.s_val );
}

char *
get_full_hostname()
{
	static char answer[ MAX_STRING ];
	char				this_host[MAX_STRING];
	struct hostent      *host_ptr;

		/* determine our own host name */
	if( gethostname(this_host,MAX_STRING) < 0 ) {
		EXCEPT( "gethostname" );
	}

		/* Look up out official host information */
	if( (host_ptr = gethostbyname(this_host)) == NULL ) {
		EXCEPT( "gethostbyname" );
	}

		/* copy out the official hostname */
	strcpy( answer, host_ptr->h_name );

	return answer;
}
