#include <stdio.h>
#include <nlist.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/proc.h>
#include <sys/param.h>

#if defined(ULTRIX42) || defined(ULTRIX43)
#include <machine/vmparam.h>
#endif

#if !defined(ULTRIX42) && !defined(SUNOS41)
#define SSYS -1
#endif

#if defined(AIX32)
#include <procinfo.h>
#endif

#include "proc_tree.h"
#include "debug.h"
#include "except.h"
static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */

#define THRESHOLD	0.60
#define BIT(n,map) (*((char *)map + (n / 8)) & (1<<(n%8)))

static struct nlist nl[] = {
	{ "_proc" },
#define X_PROC      	0
	{ "_nproc" },
#define X_NPROC     	1
	{ "_proc_bitmap" },
#define X_PROC_BITMAP   2
	{ "" },
};

static int	Kmem;			/* file descriptor for /dev/kmem */
static int	P_Tab_Offset;	/* location of process table in kmem */

static ProcInfo	*ProcTab;	/* our internal copy of the proc table */
static int	N_Procs;			/* number of procs in our table */
static int	ProcTabSize = 0;	/* # proc table slots we have allocated */

#if defined(ULTRIX42) || defined(ULTRIX43)
static int	*P_Bitmap;		/* location of process bitmap in kmem */
static int	BitmapLen;
#endif

	/* Ignore these processes and their descendents */
extern int		IgnoreList[256];
extern int		N_Ignore;

/* Pick out some field from struc proc which we don't need and can
** use to record whether we have processed this struct yet...
*/
#if defined(AIX32)
#define processed pi_adspace
#define p_pid pi_pid
#define p_ppid pi_ppid
#define p_stat pi_stat
#endif
#if defined(ULTRIX42) || defined(ULTRIX43) || defined(SUNOS41)
#define processed	p_sig
#endif


/*
** Allocate a new node for our internal process tree.  We initialize
** the state QUEUE here also.
*/
NODE *
create_node( pid, state )
int pid;
int state;
{
	NODE *answer;

	answer = (NODE *)malloc( sizeof(NODE) );
	answer->pid = pid;
	answer->state = create_queue();
	add_to_queue( answer->state, (char)state );
	answer->children = create_list();

	return answer;
}



/*
** Update our internal version of the system process tree.  We do this
** by looking through the kernel's process table in "/dev/kmem".
*/
update( tree )
NODE *tree;
{
	int		n;
	char	header[256];
	char	buf[256];
	int		i;
	NODE	*p;
	NODE	*parent;
	char	*ptr, *index();


	walk_tree( mark_untouched, tree );
	tree->touched = TRUE;

	read_proc_table();
	for( i=0; i<N_Procs; i++ ) {
		ProcTab[i].processed = FALSE;
	}

	for( i=0; i<N_Procs; i++ ) {
#if defined(ULTRIX42) || defined(ULTRIX43)
		if( !BIT(i,P_Bitmap) ) {
			continue;
		}
#endif
		if( ProcTab[i].processed ) {
			continue;
		}
		if( p = search_tree(ProcTab[i].p_pid,tree) ) {
			if( p->parent->pid != ProcTab[i].p_ppid ) {
				dprintf( D_FULLDEBUG,"Process %d has a new parent (%d)\n",
					ProcTab[i].p_pid, ProcTab[i].p_ppid );

				parent = search_tree( ProcTab[i].p_ppid, tree );
				if( parent == NULL ) {
					dprintf( D_FULLDEBUG,
								"Can't find new parent though - skipping\n" );
				} else {
					delete_from_list( p );
					add_to_list( p, parent->children );
				}
			}
			add_to_queue( p->state, state(&ProcTab[i]) );
			p->touched = TRUE;
		} else {
			insert_proc( &ProcTab[i], tree );
		}
	}

	walk_tree( delete_terminated_procs, tree );
	dprintf( D_FULLDEBUG, "Update done\n" );

}

/*
** Insert a new process into our internal process tree.
*/
void
insert_proc( proc, tree )
ProcInfo *proc;
NODE *tree;
{
	NODE	*parent_node;
	ProcInfo	*parent_proc;
	NODE	*new;
	char	*strdup();

	if( proc->processed ) {
		return;
	}

	parent_node = search_tree( proc->p_ppid, tree );
	if( parent_node == NULL ) {
		parent_proc = search_tab( proc->p_ppid );
		if( parent_proc ) {
			insert_proc( parent_proc, tree );
			parent_node = search_tree( proc->p_ppid, tree );
		} else {
			dprintf( D_FULLDEBUG, "Skipped %d\n", proc->p_pid );
			return;
		}
	}

	new = create_node( proc->p_pid, state(proc) );
	new->parent = parent_node;
	new->touched = TRUE;
#if defined(SUNOS41)
	new->type = proc->p_flag;
#endif
#if defined(ULTRIX42) || defined(ULTRIX43)
	new->type = proc->p_type;
#endif
#if defined(AIX32)
	new->type = proc->pi_flag;
#endif
	add_to_list( new, parent_node->children );
	proc->processed = TRUE;
}

/*
** Search our internal copy of the process table for a given pid.
*/
ProcInfo *
search_tab( pid )
int pid;
{
	int		i;

	for( i=0; i<N_Procs; i++ ) {
#if defined(ULTRIX42) || defined(ULTRIX43)
		if( !BIT(i,P_Bitmap) ) {
			continue;
		}
#endif
		if( ProcTab[i].p_pid == pid ) {
			return &ProcTab[i];
		}
	}
	return NULL;
}


/*
** Serach for the given pid in the tree.
*/
NODE *
search_tree( pid, tree )
int pid;
NODE *tree;
{
	NODE	*ptr;
	NODE	*answer;

	if( pid == tree->pid ) {
		return tree;
	}

	for( ptr = tree->children->next; ptr->pid != DUMMY; ptr = ptr->next ) {
		if( answer = search_tree(pid,ptr) ) {
			return answer;
		}
	}
	return NULL;
}


/*
** Add one state value to the queue.
*/
void
add_to_queue( q, val )
QUEUE *q;
int val;
{
	q->data[q->last] = (char)val;
	q->last = (q->last + 1) % QUEUE_SIZE;
	if( q->last == q->first ) {
		q->first = (q->first + 1) % QUEUE_SIZE;
	}
}

/*
** Display a state QUEUE for debugging purposes.
*/
void
display_queue( q )
QUEUE *q;
{
	int		i;

	printf( "[ " );
	for( i = q->first; i != q->last; i = (i + 1) % QUEUE_SIZE ) {
		printf( "%c ", q->data[i] );
	}
	printf( "]" );
}


/*
** Create an initial version of our internal process tree.
*/
NODE *
create_tree()
{
	NODE	*answer;

	answer = (NODE *)malloc( sizeof(NODE) );
	answer->pid = 0;
	answer->type = SSYS;
	answer->state = create_queue();
	answer->parent = answer;
	answer->children = create_list();
	return answer;
}

/*
** Display a tree node, and all its descendents.  Called with
** the root of the tree, will display the whole tree.
*/
void
display_node( n, level )
NODE *n;
int level;
{
	int		i;
	NODE	*ptr;

	for( i=0; i<level; i++ ) {
		printf( "    " );
	}

	printf( "%d", n->pid );
	display_queue( n->state );

	printf( "\n" );

	for( ptr = n->children->next; ptr->pid != DUMMY; ptr = ptr->next ) {
		display_node( ptr, level + 1 );
	}
	
}

/*
** Create a new and empty QUEUE.
*/
QUEUE *
create_queue()
{
	QUEUE	*answer;

	answer = (QUEUE *)malloc( sizeof(QUEUE) );
	answer->first = answer->last = 0;
	return answer;
}

/*
** Add a new process to a list of descendents.
*/
void
add_to_list( new, list )
NODE *new;
NODE *list;
{
	new->next = list;
	new->prev = list->prev;
	list->prev->next = new;
	list->prev = new;
}

/*
** Delete a given process from a list of descendents.
*/
void
delete_from_list( p )
NODE *p;
{
	p->next->prev = p->prev;
	p->prev->next = p->next;
}


/*
** Create a new and empty list of descendents for a process.
*/
NODE *
create_list()
{
	NODE *answer;

	answer = (NODE *)malloc( sizeof(NODE) );
	answer->pid = DUMMY;
	answer->next = answer;
	answer->prev = answer;
	answer->parent = NULL;
	answer->children = NULL;

	return answer;
}

/*
** After we have traversed our process tree with new information from
** the kernel's process table every node should be touched except
** any which represent processes that have terminated, i.e. they no
** longer appear in the kernel's process table.  Given a pointer
** to a node in our process tree, remove that node if it hasn't
** been touched.
*/
void
delete_terminated_procs( n )
NODE *n;
{
	if( n->touched == FALSE ) {
		delete_from_list( n );
	}
}

/*
** Return true if the pid is an ancestor of the process represented by
** the given node, otherwise return false.
*/
int
is_ancestor( pid, node )
int pid;
NODE *node;
{
	NODE	*ptr;

	for( ptr = node; ptr; ptr = ptr->parent ) {
		if( ptr->pid == pid ) {
			return TRUE;
		}
		if( ptr->pid == 0 ) {
			return FALSE;
		}
	}
	return FALSE;
}

/*
** Determine whether the process represented by a given node is
** "active".  This will be true if it has been in 'R' or 'D'
** state greater than a certain proportion of the time during
** its recent history as represented by the state QUEUE.
** 'R' represents runnable, and 'D' represents waiting for
** disk I/O.  We return false if we don't have enough history to
** make a reasonable determination.
*/
void
find_active( n )
NODE *n;
{
	QUEUE	*q;
	int		i;
	int		active = 0;
	int		n_values = 0;
	double	avg;

#if defined(ULTRIX42) || defined(ULTRIX43)
		/* Ignore swapper, pagedaemon, idleproc, etc. */
	if( n->type & SSYS ) {
		return;
	}
#endif

#if defined(AIX32)
		/* Ignore kproc's */
	if( n->type & SKPROC ) {
		return;
	}
#endif

	for( i=0; i<N_Ignore; i++ ) {
		if( is_ancestor(IgnoreList[i],n) ) {
			return;
		}
	}

	q = n->state;
	for( i = q->first; i != q->last; i = (i + 1) % QUEUE_SIZE ) {
		if( q->data[i] == 'R' || q->data[i] == 'D' ) {
			active += 1;
		}
		n_values += 1;
	}
	if( n_values < 4 ) {	/* not enough data to make decision */
		return;
	}

	avg = (double)active / (double)n_values;
	if( avg >= THRESHOLD ) {
		printf( "Process %d is active ", n->pid  );
		display_queue( q );
		printf( "\n" );
	}
}

void
mark_untouched( n )
NODE *n;
{
	n->touched = FALSE;
}


/*
** Apply the given function to every node in our internal process
** tree.
*/
#if defined(HAS_PROTO)

void
walk_tree( func, n )
void func(NODE *);
NODE *n;

#else

void
walk_tree( func, n )
void	(*func)();
NODE *n;

#endif
{
	NODE	*ptr;

	func( n );

	for( ptr = n->children->next; ptr->pid != DUMMY; ptr = ptr->next ) {
		walk_tree( func, ptr );
	}
	
}



/*
* Find addresses of process table in /dev/kmem.
*/
void
init_kmem()
{
    int     old_uid;

    old_uid = geteuid();
    (void)setuid( 0 );

        /* Open kmem for reading */
    if( (Kmem=open("/dev/kmem",O_RDONLY,0)) < 0 ) {
        EXCEPT( "open(/dev/kmem,O_RDONLY,0)" );
    }

    (void)seteuid( old_uid );

	/* Get the kernel's name list */
#if defined(IRIX331)						/* silicon graphics */
	(void)nlist( "/unix", nl );
	nl->n_value &= ~0x80000000;
#endif DYNIX

#if defined(HPUX9)
	(void)nlist( "/hp-ux", nl );
#endif

#if !defined(DYNIX) && !defined(AIX31) && !defined(AIX32) && !defined(IRIX331) && !defined(HPUX9) /* everybody else */
	(void)nlist( "/vmunix", nl );
#endif

		/* Find offset of kernel process table in kmem */
	read_kmem( nl[X_PROC].n_value, &P_Tab_Offset, sizeof(P_Tab_Offset) );

}

/*
** Read a structure out of /dev/kmem.
*/
read_kmem( offset, buf, len )
off_t	offset;
char	*buf;
int		len;
{

    if( lseek(Kmem,offset,0) != offset ) {
        dprintf( D_ALWAYS, "Can't seek to addr 0x%x in /dev/kmem\n", offset );
        EXCEPT( "lseek" );
    }

    if( read(Kmem,buf,len) != len ) {
        EXCEPT( "read" );
    }
}

#ifdef 0 
SetSyscalls(){}
#endif

/*
** Determine the state of the given process in a way similar to
** how "ps" would do it.
*/
char
state( p )
ProcInfo	*p;
{
		switch( p->p_stat ) {
#if defined(ULTRIX42) || defined(ULTRIX43)
			case SSLEEP:
				if( p->p_pri >= PZERO ) {
					if( p->p_slptime > MAXSLP ) {
						return 'S';
					} else {
						return 'I';
					}
				}
				if( p->p_vm & SPAGE ) {
					return 'P';
				} else {
					return 'D';
				}
#else
			case SSLEEP:
				return 'S';
#endif

#if defined(ULTRIX42) || defined(ULTRIX43)
			case SWAIT:
#endif
			case SRUN:
			case SIDL:
				return 'R';
			case SZOMB:
				return 'Z';
			case SSTOP:
				return 'T';
			default:
				return '?';
		}
}

#if defined(SUNOS41)
read_proc_table()
{
	static int	init_kmem_done;

	if( !init_kmem_done ) {
		init_kmem();
		init_kmem_done = TRUE;
	}

		/* How many processes do we have? */
	read_kmem( nl[X_NPROC].n_value, &N_Procs, sizeof(N_Procs) );

		/* allocate and read in the process table and bitmap */
	if( N_Procs > ProcTabSize ) {
		if( ProcTabSize ) {
			free( ProcTab );
			dprintf( D_FULLDEBUG, "Free'd ProcTab\n" );
		}
		ProcTab = (ProcInfo *)malloc( N_Procs * sizeof(ProcInfo) );
		dprintf( D_FULLDEBUG, "Malloc'd %d bytes for ProcTab\n",
									N_Procs * sizeof(ProcInfo) );
		ProcTabSize = N_Procs;
	}
	read_kmem( P_Tab_Offset, (char *)ProcTab, N_Procs * sizeof(ProcInfo) );
}
#endif

#if defined(ULTRIX42) || defined(ULTRIX43)
read_proc_table()
{
	static int	init_kmem_done;

	if( !init_kmem_done ) {
		init_kmem();
		init_kmem_done = TRUE;
	}

		/* How many processes do we have? */
	read_kmem( nl[X_NPROC].n_value, &N_Procs, sizeof(N_Procs) );

		/* allocate and read in the process table and bitmap */
	if( N_Procs > ProcTabSize ) {
		if( ProcTabSize ) {
			free( ProcTab );
			free( P_Bitmap );
			dprintf( D_FULLDEBUG, "Free'd ProcTab and P_Bitmap\n" );
		}
		ProcTab = (ProcInfo *)malloc( N_Procs * sizeof(ProcInfo) );
		dprintf( D_FULLDEBUG, "Malloc'd %d bytes for ProcTab\n",
									N_Procs * sizeof(ProcInfo) );
		BitmapLen = (N_Procs / 8) + 1;
		P_Bitmap = (int *)malloc( BitmapLen );
		printf( "Malloc'd %d bytes for P_Bitmap\n", BitmapLen );
		ProcTabSize = N_Procs;
	}
	read_kmem( P_Tab_Offset, (char *)ProcTab, N_Procs * sizeof(ProcInfo) );
	read_kmem( nl[X_PROC_BITMAP].n_value, P_Bitmap, BitmapLen );
}
#endif

#if defined(AIX32)
read_proc_table()
{
	int multiplier = 5;

	if( ProcTab == NULL ) {
		ProcTab = (ProcInfo *) malloc(sizeof(ProcInfo) * 100);
		if (ProcTab == NULL) {
			fprintf( stderr, "ran out of memory loading process table");
			exit( 1 );
		}
	}

	N_Procs = 0;

	while (((N_Procs = getproc(ProcTab, N_Procs, sizeof(ProcInfo))) == -1) &&
														 errno == ENOSPC) {
		N_Procs = *(long *) ProcTab;        /* num of active proc structures */
		N_Procs += (multiplier <<= 1);   /* get extra in case busy system */

		ProcTab = (ProcInfo *) realloc(ProcTab, sizeof(ProcInfo) * N_Procs);

		if (ProcTab == NULL) {
			fprintf( stderr, "ran out of memory loading process table");
			exit( 1 );
		}
	}
}
#endif
