/* 
** Copyright 1986, 1987, 1988, 1989, 1990, 1991 by the Condor Design Team
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the Condor Design Team not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the Condor
** Design Team make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE CONDOR DESIGN TEAM DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE CONDOR DESIGN TEAM BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Authors:  Allan Bricker and Michael J. Litzkow,
** 	         University of Wisconsin, Computer Sciences Dept.
** 
*/ 


#include <stdio.h>
#include <netdb.h>
#include <errno.h>
#include <pwd.h>

#if defined(IRIX331)
#define __EXTENSIONS__
#include <signal.h>
#undef __EXTENSIONS__
#else
#include <signal.h>
#endif

#include <ctype.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/param.h>
#include <sys/file.h>
#include <sys/time.h>
#include <sys/resource.h>

#if 0
#define _BSD
#include <sys/wait.h>
#else
#include <sys/wait.h>
#endif

#include <sys/stat.h>
#include <netinet/in.h>

#if defined(AIX32)
#	include "condor_fdset.h"
#endif

#undef _POSIX_SOURCE
#include "condor_types.h"

#include "debug.h"
#include "trace.h"
#include "except.h"
#include "sched.h"
#if 0 	/* now included in condor_types.h */
#include "expr.h"
#endif
#include "clib.h"

#if defined(IRIX331)
#define BADSIG SIG_ERR
#endif

static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */

CONTEXT	*create_context();
extern struct sockaddr_in   From;

extern int	errno;
extern int  HasSigchldHandler;
extern int	StarterPid;

char	*param();
void	reaper();
typedef void (*SIG_HANDLER)();
void install_sig_handler( int sig, SIG_HANDLER handler );


CONTEXT	*MachineContext;
CONTEXT	*GenericJob;
CONTEXT *JobContext;

int	FrgnJobsRunning;

char	ThisHost[512];
char	XSockName[512];

char	*Log;
int		PollingFrequency;
int		PollsPerUpdate;
char	*Execute;
char	*CollectorHost;
char	*AlternateCollectorHost;
char	*CondorAdministrator;
int		Foreground;
int		Termlog;
int		Memory;
char	*Spool;
char	*KbdDev;
char	*MouseDev;
int		State;
char	*Mail;

/*
  Allow a primary version of the condor_starter, with up to 10 alternate
  versions.  These are to be listed in the config files as
	STARTER
	ALTERNATE_STARTER_0
	ALTERNATE_STARTER_1
	ALTERNATE_STARTER_2
	...
	ALTERNATE_STARTER_9
  Only the first entry is required.  We select the starter based on the
  command number given by the condor_shadow.  Command "SCHED_VERS+11"
  is the normal "START_FRGN_JOB" request as before.  Commands
  "SCHED_VERS+70" - "SCHED_VERS+79" are also start job requests, but
  select their respective alternate versions of the starter.  Any of
  these requests can be refused if no such starter is defined in the
  config files.
*/
#define STARTER_0 (SCHED_VERS+ALT_STARTER_BASE+0)
#define STARTER_1 (SCHED_VERS+ALT_STARTER_BASE+1)
#define STARTER_2 (SCHED_VERS+ALT_STARTER_BASE+2)
#define STARTER_3 (SCHED_VERS+ALT_STARTER_BASE+3)
#define STARTER_4 (SCHED_VERS+ALT_STARTER_BASE+4)
#define STARTER_5 (SCHED_VERS+ALT_STARTER_BASE+5)
#define STARTER_6 (SCHED_VERS+ALT_STARTER_BASE+6)
#define STARTER_7 (SCHED_VERS+ALT_STARTER_BASE+7)
#define STARTER_8 (SCHED_VERS+ALT_STARTER_BASE+8)
#define STARTER_9 (SCHED_VERS+ALT_STARTER_BASE+9)
char	*PrimaryStarter;
char	*AlternateStarter[10];
char	*Starter;

int		On = 1;

#ifdef vax
struct linger linger = { 0, 0 };	/* Don't linger */
#endif vax

int		ConnectionSock;
int		UdpSock;	/* Send datagrams to the collector */
int		UdpSockAlternate;	/* Send datagrams to the alternate collector */
int		UdpIn;		/* Get datagrams from the keyboard daemon */

extern int	Terse;
int	LastTimeout;

char *MyName;

usage( name )
char	*name;
{
	dprintf( D_ALWAYS, "Usage: %s [-f] [-t]\n", name );
	exit( 1 );
}

char	GenericOwner[] = "Owner = \"nobody\"";


main( argc, argv)
int		argc;
char	*argv[];
{
	int		count;
	fd_set	readfds;
	struct timeval	timer;
	char	**ptr;
	void	sigint_handler(), sighup_handler();
	int		i;

#ifdef NFSFIX
	if( getuid() != 0 ) {
		EXCEPT("The start daemon must be run as root");
	}

	/* We must be run as root, but we want to be condor to write to logs. */
	set_condor_euid(__FILE__,__LINE__);
#endif NFSFIX


		/* Read the configuration file */
	MachineContext = create_context();
	MyName = *argv;
	config( MyName, MachineContext );
	init_params();
	Terse = 1;

	/*
	** Check and correct(if necessary) the permissions on the execute directory.
	** Has to be done after init_params so we know the pathname.
	*/
	check_perms();


		/* Set up a "generic" job which we can use to evaluate whether we
		are "idle" or not. */
	GenericJob = create_context();
    store_stmt( scan(GenericOwner), GenericJob);

	cleanup_execute_dir();

	if( argc > 3 ) {
		usage( argv[0] );
	}
	for( ptr=argv+1; *ptr; ptr++ ) {
		if( ptr[0][0] != '-' ) {
			usage( argv[0] );
		}
		switch( ptr[0][1] ) {
			case 'f':
				Foreground++;
				break;
			case 't':
				Termlog++;
				break;
			default:
				usage( argv[0] );
		}
	}

		/* This is so if we dump core it'll go in the log directory */
	if( chdir(Log) < 0 ) {
		EXCEPT( "chdir to log directory <%s>", Log );
	}

		/* Arrange to run in background */
	if( !Foreground ) {
		if( fork() )
			exit( 0 );
	}

		/* Set up logging */
	dprintf_config( "STARTD", 2 );

	dprintf( D_ALWAYS, "**************************************************\n" );
	dprintf( D_ALWAYS, "***          CONDOR_STARTD STARTING UP         ***\n" );
	dprintf( D_ALWAYS, "**************************************************\n" );
	dprintf( D_ALWAYS, "\n" );

	dprintf( D_ALWAYS, "PRIMARY_STARTER = \"%s\"\n", PrimaryStarter );
	for( i=0; i<10; i++ ) {
		dprintf( D_ALWAYS,
			"ALTERNATE_STARTER_%d = \"%s\"\n", i, AlternateStarter[i]
		);
	}

#ifndef NFSFIX
	if( getuid() != 0 ) {
		EXCEPT("The start daemon must be run as root");
	}
#endif NFSFIX

	init_fs_info();
	install_sig_handler( SIGINT, sigint_handler );
	install_sig_handler( SIGHUP, sighup_handler );
	install_sig_handler( SIGPIPE, SIG_IGN );

	init_environ();

	ConnectionSock = init_connection_sock( "condor_startd", START_PORT );
	UdpIn = init_udp_sock( "condor_startd", START_UDP_PORT );
	UdpSock = udp_connect( CollectorHost, COLLECTOR_UDP_PORT );
	if ( AlternateCollectorHost != NULL )
	 UdpSockAlternate = udp_connect( AlternateCollectorHost, COLLECTOR_UDP_PORT );

	install_sig_handler( SIGCHLD, reaper );
	HasSigchldHandler = TRUE;

	if( gethostname(ThisHost,sizeof ThisHost) < 0 ) {
		EXCEPT( "gethostname" );
	}
	(void)sprintf( XSockName, "%s:0", ThisHost );

   get_k_vars();   /* prepare to read kernel's process table */

	LastTimeout = (int)time( (time_t *)0 );
	timer.tv_usec = 0;

	change_states( NO_JOB );
	timeout();
	FD_ZERO( &readfds );

	/*
	clearalloc( "/tmp/mike", 0 );
	*/


	for(;;) {
		/*
		showalloc( "Top of loop" );
		dprintf( D_ALWAYS, "BRK = 0x%x\n", sbrk(0) );
		*/

		FD_SET( ConnectionSock, &readfds );
		FD_SET( UdpIn, &readfds );
		timer.tv_sec = PollingFrequency -
									( (int)time((time_t *)0) - LastTimeout );
		if( timer.tv_sec < 0 ) {
			timer.tv_sec = 0;
		}

#if defined(AIX31) || defined(AIX32)
		errno = EINTR;	/* Shouldn't need to do this... */
#endif
		count = select(FD_SETSIZE, (int *)&readfds, (int *)0, (int *)0,
												(struct timeval *)&timer );
		

		if( count < 0 ) {
			if( errno == EINTR ) {
				dprintf( D_FULLDEBUG, "select was interrupted - retrying...\n");
				continue;
			} else {
				EXCEPT( "select(FD_SETSIZE,0%o,0,0,%d sec)",
												readfds, timer.tv_sec );
			}
		}

		(void)sigblock( sigmask(SIGCHLD) ); 	/* Block out SIGCHLD */
		if( NFDS(count) == 0 ) {
			timeout();
		}
		if( FD_ISSET(ConnectionSock,&readfds) ) {
			accept_connection();
		}
		if( FD_ISSET(UdpIn,&readfds) ) {
			accept_datagram();
		}
		(void)sigsetmask( 0 );			/* Allow SIGCHLD */
	}
}

accept_connection()
{
	struct sockaddr_in	from;
	int		len;
	int		fd;
	XDR		xdr, *xdrs, *xdr_Init();

	len = sizeof from;
	bzero( (char *)&from, sizeof from );
	fd = accept( ConnectionSock, (struct sockaddr *)&from, &len );

	if( fd < 0 && errno != EINTR ) {
		EXCEPT( "accept" );
	}

	if( fd >= 0 ) {
		xdrs = xdr_Init( &fd, &xdr );
		do_command( xdrs, &from );
		xdr_destroy( xdrs );
		(void)close( fd );
	}
}

accept_datagram()
{
    XDR     xdr, *xdrs, *xdr_Udp_Init();

    xdrs = xdr_Udp_Init( &UdpIn, &xdr );

	do_command( xdrs, &From );

    xdr_destroy( xdrs );
}


init_connection_sock( service, port )
char	*service;
int		port;
{
	struct sockaddr_in	sin;
	struct servent *servp;
	int		sock;
	int		status;

	bzero( (char *)&sin, sizeof sin );
	servp = getservbyname(service, "tcp");
	if( servp ) {
		sin.sin_port = htons( (u_short)servp->s_port );
	} else {
		sin.sin_port = htons( (u_short)port );
	}

	if( (sock=socket(AF_INET,SOCK_STREAM,0)) < 0 ) {
		EXCEPT( "socket" );
	}

	if( setsockopt(sock,SOL_SOCKET,SO_REUSEADDR,(caddr_t *)&On,sizeof(On))< 0) {
		EXCEPT( "setsockopt" );
	}

#ifdef vax
	if( setsockopt(sock,SOL_SOCKET,SO_LINGER,&linger,sizeof(linger)) < 0 ) {
		EXCEPT( "setsockopt" );
	}
#endif vax

	if( (status = bind(sock,(struct sockaddr *)&sin,sizeof(sin))) < 0 ) {
		dprintf( D_ALWAYS, "Bind returns %d, errno = %d\n", status, errno );
		if( errno == EADDRINUSE ) {
			EXCEPT( "CONDOR_STARTD ALREADY RUNNING" );
		} else {
			EXCEPT( "bind" );
		}
	}

	if( listen(sock,5) < 0 ) {
		EXCEPT( "listen" );
	}

	return sock;
}

#if 0
void
reaper()
{
	int		pid;
	union wait	status;
	struct rusage rusage;

	while( (pid = wait3(&status,WNOHANG|WUNTRACED,&rusage)) > 0 )  {

		if( pid != StarterPid ) {
			continue;
		}

		if( WIFSTOPPED(status) ) {
			dprintf( D_ALWAYS, "Starter %d has stopped\n", pid );
			return;
		}
		if( WIFSIGNALED(status) ) {
			dprintf( D_ALWAYS, "Starter %d died on signal %d, coredump = %d\n",
									pid, status.w_termsig, status.w_coredump );
		} else {
			dprintf( D_ALWAYS, "Starter %d exited with status %d\n",
											pid, status.w_retcode );
		}
		job_dies();
	}
}
#else
void
reaper()
{
	int		pid;
	int	status;

	while( (pid = waitpid(-1,&status,WNOHANG|WUNTRACED)) > 0 )  {

		if( pid != StarterPid ) {
			continue;
		}

		if( WIFSTOPPED(status) ) {
			dprintf( D_ALWAYS, "Starter %d has stopped\n", pid );
			return;
		}
		if( WIFSIGNALED(status) ) {
			dprintf( D_ALWAYS, "Starter %d died on signal %d\n",
									pid, WTERMSIG(status) );
		} else {
			dprintf( D_ALWAYS, "Starter %d exited with status %d\n",
											pid, WEXITSTATUS(status) );
		}
		job_dies();
	}
}
#endif

/*
** Somebody has connected to our socket with a request.  Read the request
** and handle it.
*/
do_command( xdrs, from )
XDR		*xdrs;
struct sockaddr_in	*from;
{
	int		cmd;
	int		index;
	struct hostent  *hp, *gethostbyaddr();

		/* Read the request */
	xdrs->x_op = XDR_DECODE;
	if( !xdr_int(xdrs,&cmd) ) {
		dprintf( D_ALWAYS, "Can't read command\n" );
		return;
	}

	dprintf( D_FULLDEBUG, "Got command %d.\n", cmd );

	switch( cmd ) {
		case STARTER_0:
		case STARTER_1:
		case STARTER_2:
		case STARTER_3:
		case STARTER_4:
		case STARTER_5:
		case STARTER_6:
		case STARTER_7:
		case STARTER_8:
		case STARTER_9:
				/* Special command - set up to use an alternate starter */
			dprintf( D_ALWAYS, "Got start_job_request\n" );
			display_from( from );
			index = cmd - SCHED_VERS - ALT_STARTER_BASE;
			Starter = AlternateStarter[index];
			dprintf( D_ALWAYS, "Using starter \"%s\"\n", Starter );
			start_job_request( xdrs, from );
			break;
		case START_FRGN_JOB:	/* Request from foreign shadow to start a job */
			dprintf( D_ALWAYS, "Got start_job_request\n" );
			display_from( from );
			Starter = PrimaryStarter;	/* Restore the primary starter */
			dprintf( D_ALWAYS, "Using starter \"%s\"\n", Starter );
			start_job_request( xdrs, from );
			break;
		case CKPT_FRGN_JOB:
			dprintf( D_ALWAYS, "Got ckpt_order\n" );
			display_from( from );
			ckpt_order();
			break;
		case KILL_FRGN_JOB:
			dprintf( D_ALWAYS, "Got kill_job\n" );
			display_from( from );
			kill_job( OK );
			break;
		case X_EVENT_NOTIFICATION:
			x_event();
			break;
		case REQ_NEW_PROC:
			/* Request from foreign shadow to run a 2nd or more process here */
			display_from( from );
			new_proc_order();
			break;
		default:
			dprintf( D_ALWAYS, "Got unknown command (%d).\n", cmd );
			display_from( from );
	}
}

display_from( from )
struct sockaddr_in	*from;
{
	struct hostent  *hp, *gethostbyaddr();

	if( (hp=gethostbyaddr((char *)&from->sin_addr,
				sizeof(struct in_addr), from->sin_family)) == NULL ) {
		dprintf( D_ALWAYS, "From (%s), port %d\n",
			inet_ntoa(from->sin_addr), from->sin_port );
	} else {
		dprintf( D_ALWAYS, "From %s, port %d\n",
										hp->h_name, from->sin_port );
	}
}

#ifdef 0
SetSyscalls(){}
#endif

init_params()
{
	char	*pval;
	char	*tmp;

	init_starters();

	Log = param( "LOG" );
	if( Log == NULL )  {
		EXCEPT( "No log directory specified in config file\n" );
	}

	Execute = param( "EXECUTE" );
	if( Execute == NULL ) {
		EXCEPT( "No Execute file specified in config file\n" );
	}

	CollectorHost = param( "COLLECTOR_HOST" );
	if( CollectorHost == NULL ) {
		EXCEPT( "No Collector host specified in config file\n" );
	}

	AlternateCollectorHost = param( "ALTERNATE_COLLECTOR_HOST" );

	tmp = param( "POLLING_FREQUENCY" );
	if( tmp == NULL ) {
		PollingFrequency = 30;
	} else {
		PollingFrequency = atoi( tmp );
		free( tmp );
	}

	tmp = param( "POLLS_PER_UPDATE" );
	if( tmp == NULL ) {
		PollsPerUpdate = 4;
	} else {
		PollsPerUpdate = atoi( tmp );
		free( tmp );
	}

	if( (tmp=param("STARTD_DEBUG")) == NULL ) {
		EXCEPT( "STARTD_DEBUG not defined in config file\n" );
	}
	free( tmp );
	Foreground = boolean( "STARTD_DEBUG", "Foreground" );

	if( (CondorAdministrator = param("CONDOR_ADMIN")) == NULL ) {
		EXCEPT( "CONDOR_ADMIN not specified in config file" );
	}

    pval = param("MEMORY");
    if( pval != NULL ) {
        Memory = atoi( pval );
		free( pval );
    }
    if( Memory <= 0 || Memory >= 500 ) {
        Memory =  DEFAULT_MEMORY;
    }

    if( (Spool = param("SPOOL")) == NULL ) {
        EXCEPT( "SPOOL not specified in config file" );
    }

	KbdDev = param( "KBD_DEVICE" );
	MouseDev = param( "MOUSE_DEVICE" );

	if( (Mail = param("MAIL")) == NULL ) {
		EXCEPT( "MAIL not specified in config file" );
	}

}

init_starters()
{
	char	*pval;
	char	*tmp;
	int		i;
	char	buf[ 1024 ];

	PrimaryStarter = param( "STARTER" );
	if( PrimaryStarter == NULL ) {
		EXCEPT( "No Starter file specified in config file\n" );
	}
	Starter = PrimaryStarter;

	for( i=0; i<10; i++ ) {
		sprintf( buf, "ALTERNATE_STARTER_%d", i );
		AlternateStarter[i] = param(buf);
	}

}


void
sigint_handler()
{
	dprintf( D_ALWAYS, "Killed by SIGINT\n" );
	exit( 0 );
}

void
sighup_handler()
{
	dprintf( D_ALWAYS, "Re reading config file\n" );

	free_context( MachineContext );
	MachineContext = create_context();
	config( MyName, MachineContext );
	init_params();
	init_fs_info();
}

cleanup_execute_dir()
{
	 char *execute, *rindex();
	 char buf[ 1024 ];

	execute = rindex(Execute, '/');
	if( execute == NULL ) {
		execute = Execute;
	} else {
		execute++;
	}

	if( strcmp("execute", execute) != 0 ) {
		EXCEPT("EXECUTE parameter (%s) must end with 'execute'.\n", Execute);
	}

			/* Note: Unfortunately, if the starter should dump core, this
			   will remove it. */
	(void)sprintf(buf, "rm -rf %s/condor_exec* %s/dir_*", Execute, Execute );


	(void)system(buf);
}

init_udp_sock( service, port )
char    *service;
int     port;
{
    struct sockaddr_in  sin;
    struct servent *servp;
    int     sock;

    bzero( (char *)&sin, sizeof sin );
    servp = getservbyname(service, "udp");
    if( servp ) {
        sin.sin_port = htons( (u_short)servp->s_port );
    } else {
        sin.sin_port = htons( (u_short)port );
    }

    if( (sock=socket(AF_INET,SOCK_DGRAM,0)) < 0 ) {
        EXCEPT( "socket" );
    }

    if( bind(sock,(struct sockaddr *)&sin,sizeof(sin)) < 0 ) {
        if( errno == EADDRINUSE ) {
            EXCEPT( "CONDOR_STARTD ALREADY RUNNING" );
        } else {
            EXCEPT( "bind" );
        }
    }

    return sock;
}

/*
** Make sure we have "/etc" and "/usr/etc" in our PATH so we can find
** things like "pstat" later.
*/
extern char	**environ;
char	*DefaultPath = "PATH=/bin:/usr/bin:/usr/ucb:/etc:/usr/etc";
init_environ()
{
	char	**src, **dst;
	static char buf[1024];
	static char	*new_environ[512];
	int		path_set = 0;

	for( src=environ, dst=new_environ; *src; src++, dst++ ) {
		if( strncmp("PATH",*src,4) == MATCH ) {
			sprintf( buf, "%s:/etc:/usr/etc", *src  );
			*dst = buf;
			path_set = 1;
		} else {
			*dst = *src;
		}
	}
	if( !path_set ) {
		*dst = DefaultPath;
		dst++;
	}
	*dst = NULL;
	environ = new_environ;

	dprintf( D_FULLDEBUG, "New environment\n" );
	for( src=environ; *src; src++ ) {
		dprintf( D_FULLDEBUG, "%s\n", *src );
	}
}

#if defined(BSD43) || defined(DYNIX)
#define mode_t unsigned short
#endif

/*
** The permissions on the execute directory should be read, write, and execute
** by anyone and everyone.
*/
#define PERMMASK 00777
check_perms()
{
	struct stat statBuf;
	mode_t mode;

	/*
	** Find the permission of the file.
	*/
	if(stat(Execute, &statBuf) == -1) {
		EXCEPT("stat");
	}

	/*
	** Check the permission of the file.  Only concerned with the bits
	** applicable to PERMMASK.
	*/
    if( (statBuf.st_mode & PERMMASK) != PERMMASK ) {
		dprintf(D_FULLDEBUG, "Changing permission on execute directory.\n");
		mode = statBuf.st_mode | PERMMASK;
		if(chmod(Execute, mode) == -1) {
			EXCEPT("chmod");
		}
	}
}
