/* 
** Copyright 1986, 1987, 1988, 1989, 1990, 1991 by the Condor Design Team
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the Condor Design Team not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the Condor
** Design Team make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE CONDOR DESIGN TEAM DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE CONDOR DESIGN TEAM BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Authors:  Allan Bricker and Michael J. Litzkow,
** 	         University of Wisconsin, Computer Sciences Dept.
** 
*/ 


#include <rpc/types.h>
#include <rpc/xdr.h>
#include  <stdio.h>
#include "sched.h"
#include "debug.h"

char *strdup( const char * );

xdr_ports( xdrs, ports )
XDR		*xdrs;
PORTS	*ports;
{
	return xdr_int(xdrs,&ports->port1) && xdr_int(xdrs,&ports->port2);
}


/*
**	Routine to transfer the initial information from the startd to
**  the shadow.
**			Author : Dhrubajyoti Borthakur
*/

int
xdr_startd_info(xdrs, startdRec)
XDR		*xdrs;
StartdRec	*startdRec;
{
	char	*ptr = NULL;
	if ( !xdr_int(xdrs, &startdRec->version_num)) return FALSE;

	dprintf(D_ALWAYS, "Version = %d\n", startdRec->version_num);
	if ( startdRec->version_num >=0 )
	{
		/* we are talking with a startd of an old version which sends just
		   two port numbers */
		startdRec->ports.port1 = startdRec->version_num;
		if ( !xdr_int(xdrs, &(startdRec->ports.port2))) return FALSE;
		dprintf(D_ALWAYS, "Port2 = %d\n", startdRec->ports.port2);
		return TRUE;
	}
	if ( !xdr_ports(xdrs, &startdRec->ports)) return FALSE;
	if ( !xdr_int(xdrs, &startdRec->ip_addr)) return FALSE;
	dprintf(D_ALWAYS, "IP-addr = 0x%x\n", startdRec->ip_addr);

	if ( xdrs->x_op == XDR_ENCODE ) 
	{
		ptr =  startdRec->server_name;
		if ( !xdr_mywrapstring(xdrs, &ptr)) return FALSE;
		dprintf(D_ALWAYS, "XDR:Send server_name = %s\n", ptr);
	}
	else if ( xdrs->x_op == XDR_DECODE) 
	{
		ptr = NULL;
		if ( !xdr_mywrapstring(xdrs, &ptr)) return FALSE;
		if ( ptr ) startdRec->server_name = strdup(ptr);
		dprintf(D_ALWAYS, "XDR:Received server_name = %s\n", ptr);
	}
	return TRUE;
}

