/* 
** Copyright 1986, 1987, 1988, 1989, 1990, 1991 by the Condor Design Team
** 
** Permission to use, copy, modify, and distribute this software and its
** documentation for any purpose and without fee is hereby granted,
** provided that the above copyright notice appear in all copies and that
** both that copyright notice and this permission notice appear in
** supporting documentation, and that the names of the University of
** Wisconsin and the Condor Design Team not be used in advertising or
** publicity pertaining to distribution of the software without specific,
** written prior permission.  The University of Wisconsin and the Condor
** Design Team make no representations about the suitability of this
** software for any purpose.  It is provided "as is" without express
** or implied warranty.
** 
** THE UNIVERSITY OF WISCONSIN AND THE CONDOR DESIGN TEAM DISCLAIM ALL
** WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING ALL IMPLIED WARRANTIES
** OF MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE UNIVERSITY OF
** WISCONSIN OR THE CONDOR DESIGN TEAM BE LIABLE FOR ANY SPECIAL, INDIRECT
** OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS
** OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
** OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE
** OR PERFORMANCE OF THIS SOFTWARE.
** 
** Authors:  Allan Bricker and Michael J. Litzkow,
** 	         University of Wisconsin, Computer Sciences Dept.
** 
*/ 


#include <stdlib.h>
#include <sys/types.h>
#include <sys/param.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <rpc/types.h>
#include <rpc/xdr.h>

#include "trace.h"
#include "except.h"
#include "xdr_lib.h"

#ifndef NEW_PROC
#define NEW_PROC
#endif
#include "proc.h"

#include "debug.h"

#ifndef LINT
static char *_FileName_ = __FILE__;		/* Used by EXCEPT (see except.h)     */
#endif

/*
** XDR routine for a PROC.
*/
xdr_proc( xdrs, proc )
XDR		*xdrs;
PROC	*proc;
{

#if defined(HPUX9)
	if( proc == NULL ) {
		return FALSE;
	}
#endif

	if( !xdr_int(xdrs,&proc->version_num) )
		return FALSE;
	switch( proc->version_num ) {
		case 2:
			return xdr_proc_vers2( xdrs, proc );
			break;
#if defined(NEW_PROC)
		case 3:
			return xdr_proc_vers3( xdrs, proc );
			break;
#endif
		default:
			EXCEPT( "Incorrect PROC version number (%d)", proc->version_num );
	}
}

#if defined(NEW_PROC)
/*
** XDR routine for a version 3 PROC struct.
*/
xdr_proc_vers3( xdrs, proc )
XDR		*xdrs;
PROC	*proc;
{
	int		i;

	if( !xdr_proc_id(xdrs,&proc->id) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->universe) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->checkpoint) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->remote_syscalls) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->owner,256) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->q_date) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->completion_date) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->status) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->prio) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->notification) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->image_size) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->env,MAXPATHLEN) )
		return FALSE;

	if( !xdr_int(xdrs,&proc->n_cmds) ) {
		return FALSE;
	}

	if( xdrs->x_op == XDR_DECODE ) {
		proc->cmd = (char **)calloc( proc->n_cmds, sizeof(char *) );
		proc->args = (char **)calloc( proc->n_cmds, sizeof(char *) );
		proc->in = (char **)calloc( proc->n_cmds, sizeof(char *) );
		proc->out = (char **)calloc( proc->n_cmds, sizeof(char *) );
		proc->err = (char **)calloc( proc->n_cmds, sizeof(char *) );
		proc->remote_usage = (struct rusage *)
						calloc( proc->n_cmds, sizeof(struct rusage) );
		proc->exit_status = (int *)
						calloc( proc->n_cmds, sizeof(int) );
	}

	for( i=0; i<proc->n_cmds; i++ ) {
		if( !xdr_string(xdrs,&proc->cmd[i],MAXPATHLEN) )
			return FALSE;
		if( !xdr_string(xdrs,&proc->args[i],MAXPATHLEN) )
			return FALSE;
		if( !xdr_string(xdrs,&proc->in[i],MAXPATHLEN) )
			return FALSE;
		if( !xdr_string(xdrs,&proc->out[i],MAXPATHLEN) )
			return FALSE;
		if( !xdr_string(xdrs,&proc->err[i],MAXPATHLEN) )
			return FALSE;
		if( !xdr_rusage(xdrs,&proc->remote_usage[i]) )
			return FALSE;
		if( !xdr_int(xdrs,&proc->exit_status[i]) )
			return FALSE;
	}

	if( xdrs->x_op == XDR_FREE ) {
		free( proc->cmd );
		free( proc->args );
		free( proc->in );
		free( proc->out );
		free( proc->err );
		free( proc->remote_usage );
		free( proc->exit_status );
	}

	if( !xdr_int(xdrs,&proc->n_pipes) )
		return FALSE;

	if( proc->n_pipes > 0 && xdrs->x_op == XDR_DECODE ) {
		proc->pipe = (P_DESC *)calloc( proc->n_pipes, sizeof(P_DESC) );
	}

	for( i=0; i<proc->n_pipes; i++ ) {
		if( !xdr_pipe_desc(xdrs,&proc->pipe[i]) )
			return FALSE;
	}

	if( proc->n_pipes > 0 && xdrs->x_op == XDR_FREE ) {
		free( proc->pipe );
	}

	if ( !xdr_int(xdrs, &(proc->min_needed)) ) {
		return FALSE;
	}

	if ( !xdr_int(xdrs, &(proc->max_needed)) ) {
		return FALSE;
	}

	if( !xdr_string(xdrs,&proc->rootdir,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->iwd,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->requirements,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->preferences,MAXPATHLEN) )
		return FALSE;
	if( !xdr_rusage(xdrs,&proc->local_usage) )
		return FALSE;
	return TRUE;
}
#endif

/*
** XDR routine for a version 2 PROC struct.
*/
xdr_proc_vers2( xdrs, proc )
XDR		*xdrs;
V2_PROC	*proc;
{

	if( !xdr_proc_id(xdrs,&proc->id) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->owner,256) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->q_date) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->completion_date) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->status) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->prio) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->notification) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->image_size) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->cmd,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->args,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->env,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->in,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->out,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->err,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->rootdir,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->iwd,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->requirements,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->preferences,MAXPATHLEN) )
		return FALSE;
	if( !xdr_rusage(xdrs,&proc->local_usage) )
		return FALSE;
	if( !xdr_rusage(xdrs,&proc->remote_usage) )
		return FALSE;
	return TRUE;
}

xdr_free_proc( proc )
PROC	*proc;
{
	XDR		xdrs;

	xdrs.x_op = XDR_FREE;
	xdr_proc( &xdrs, proc );
}

/*
** XDR routine for a PROC_ID.
*/
xdr_proc_id( xdrs, id )
XDR		*xdrs;
PROC_ID	*id;
{
	return xdr_int(xdrs,&id->cluster) && xdr_int(xdrs,&id->proc);
}

/*
** XDR routine for a PROC.
*/
xdr_old_proc( xdrs, proc )
XDR		*xdrs;
OLD_PROC	*proc;
{
	if( xdrs->x_op == XDR_FREE ) {
		xdr_old_free_proc( proc );
		return TRUE;
	}

	if( !xdr_proc_id(xdrs,&proc->id) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->owner,256) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->q_date) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->status) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->prio) )
		return FALSE;
	if( !xdr_int(xdrs,&proc->notification) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->cmd,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->args,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->env,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->in,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->out,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->err,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->rootdir,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->iwd,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->requirements,MAXPATHLEN) )
		return FALSE;
	if( !xdr_string(xdrs,&proc->preferences,MAXPATHLEN) )
		return FALSE;
	if( !xdr_rusage(xdrs,&proc->local_usage) )
		return FALSE;
	if( !xdr_rusage(xdrs,&proc->remote_usage) )
		return FALSE;
	return TRUE;
}

xdr_old_free_proc( proc )
OLD_PROC	*proc;
{
	if( proc->owner ) {
		FREE( proc->owner );
		proc->owner = NULL;
	}

	if( proc->cmd ) {
		FREE( proc->cmd );
		proc->cmd = NULL;
	}

	if( proc->args ) {
		FREE( proc->args );
		proc->args = NULL;
	}

	if( proc->env ) {
		FREE( proc->env );
		proc->env = NULL;
	}

	if( proc->in ) {
		FREE( proc->in );
		proc->in = NULL;
	}

	if( proc->out ) {
		FREE( proc->out );
		proc->out = NULL;
	}

	if( proc->err ) {
		FREE( proc->err );
		proc->err = NULL;
	}

	if( proc->rootdir ) {
		FREE( proc->rootdir );
		proc->rootdir = NULL;
	}

	if( proc->iwd ) {
		FREE( proc->iwd );
		proc->iwd = NULL;
	}

	if( proc->requirements ) {
		FREE( proc->requirements );
		proc->requirements = NULL;
	}

	if( proc->preferences ) {
		FREE( proc->preferences );
		proc->preferences = NULL;
	}
}

xdr_pipe_desc( xdrs, pipe )
XDR		*xdrs;
P_DESC	*pipe;
{
	if( !xdr_int(xdrs,&pipe->writer) )
		return FALSE;
	if( !xdr_int(xdrs,&pipe->reader) )
		return FALSE;
	if( !xdr_string(xdrs,&pipe->name,MAXPATHLEN) )
		return FALSE;
	return TRUE;
}
